# -*- coding: utf-8 -*-
#
# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
    Parse functions for Portuguese (PT-PT)

    TODO: numbers greater than 999999
    TODO: date time pt
"""

from datetime import datetime

from dateutil.relativedelta import relativedelta

from tnalagmes.lang.parse_common import is_numeric, look_for_fractions

# Undefined articles ["um", "uma", "uns", "umas"] can not be supressed,
# in PT, "um cavalo" means "a horse" or "one horse".
pt_articles = ["o", "a", "os", "as"]

pt_numbers = {
    "zero": 0,
    "um": 1,
    "uma": 1,
    "uns": 1,
    "umas": 1,
    "primeiro": 1,
    "segundo": 2,
    "terceiro": 3,
    "dois": 2,
    "duas": 2,
    "tres": 3,
    u"três": 3,
    "quatro": 4,
    "cinco": 5,
    "seis": 6,
    "sete": 7,
    "oito": 8,
    "nove": 9,
    "dez": 10,
    "onze": 11,
    "doze": 12,
    "treze": 13,
    "catorze": 14,
    "quinze": 15,
    "dezasseis": 16,
    "dezassete": 17,
    "dezoito": 18,
    "dezanove": 19,
    "vinte": 20,
    "trinta": 30,
    "quarenta": 40,
    "cinquenta": 50,
    "sessenta": 60,
    "setenta": 70,
    "oitenta": 80,
    "noventa": 90,
    "cem": 100,
    "cento": 100,
    "duzentos": 200,
    "duzentas": 200,
    "trezentos": 300,
    "trezentas": 300,
    "quatrocentos": 400,
    "quatrocentas": 400,
    "quinhentos": 500,
    "quinhentas": 500,
    "seiscentos": 600,
    "seiscentas": 600,
    "setecentos": 700,
    "setecentas": 700,
    "oitocentos": 800,
    "oitocentas": 800,
    "novecentos": 900,
    "novecentas": 900,
    "mil": 1000,
    u"milhï¿½o": 1000000}


def isFractional_pt(input_str):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        text (str): the string to check if fractional
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    if input_str.endswith('s', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "fifths"

    aFrac = ["meio", u"terço", "quarto", "quinto", "sexto",
             "setimo", "oitavo", "nono", u"décimo"]

    if input_str.lower() in aFrac:
        return 1.0 / (aFrac.index(input_str) + 2)
    if input_str == u"vigésimo":
        return 1.0 / 20
    if input_str == u"trigésimo":
        return 1.0 / 30
    if input_str == u"centésimo":
        return 1.0 / 100
    if input_str == u"milésimo":
        return 1.0 / 1000
    if (input_str == u"sétimo" or input_str == "septimo" or
            input_str == u"séptimo"):
        return 1.0 / 7

    return False


def extractnumber_pt(text):
    """
    This function prepares the given text for parsing by making
    numbers consistent, getting rid of contractions, etc.
    Args:
        text (str): the string to normalize
    Returns:
        (int) or (float): The value of extracted number

    """
    aWords = text.split()
    count = 0
    result = None
    while count < len(aWords):
        val = 0
        word = aWords[count]
        next_next_word = None
        if count + 1 < len(aWords):
            next_word = aWords[count + 1]
            if count + 2 < len(aWords):
                next_next_word = aWords[count + 2]
        else:
            next_word = None

        # is current word a number?
        if word in pt_numbers:
            val = pt_numbers[word]
        elif word.isdigit():  # doesn't work with decimals
            val = int(word)
        elif is_numeric(word):
            val = float(word)
        elif isFractional_pt(word):
            if not result:
                result = 1
            result = result * isFractional_pt(word)
            count += 1
            continue

        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            # if (len(aPieces) == 2 and is_numeric(aPieces[0])
            #   and is_numeric(aPieces[1])):
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])

        if val:
            if result is None:
                result = 0
            # handle fractions
            if next_word != "avos":
                result += val
            else:
                result = float(result) / float(val)

        if next_word is None:
            break

        # number word and fraction
        ands = ["e"]
        if next_word in ands:
            zeros = 0
            if result is None:
                count += 1
                continue
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "

            afterAndVal = extractnumber_pt(newText[:-1])
            if afterAndVal:
                if result < afterAndVal or result < 20:
                    while afterAndVal > 1:
                        afterAndVal = afterAndVal / 10.0
                    for word in newWords:
                        if word == "zero" or word == "0":
                            zeros += 1
                        else:
                            break
                for _ in range(0, zeros):
                    afterAndVal = afterAndVal / 10.0
                result += afterAndVal
                break
        elif next_next_word is not None:
            if next_next_word in ands:
                newWords = aWords[count + 3:]
                newText = ""
                for word in newWords:
                    newText += word + " "
                afterAndVal = extractnumber_pt(newText[:-1])
                if afterAndVal:
                    if result is None:
                        result = 0
                    result += afterAndVal
                    break

        decimals = ["ponto", "virgula", u"vï¿½rgula", ".", ","]
        if next_word in decimals:
            zeros = 0
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "
            for word in newWords:
                if word == "zero" or word == "0":
                    zeros += 1
                else:
                    break
            afterDotVal = str(extractnumber_pt(newText[:-1]))
            afterDotVal = zeros * "0" + afterDotVal
            result = float(str(result) + "." + afterDotVal)
            break
        count += 1

    if result is None:
        return False

    # Return the $str with the number related words removed
    # (now empty strings, so strlen == 0)
    # aWords = [word for word in aWords if len(word) > 0]
    # text = ' '.join(aWords)
    if "." in str(result):
        integer, dec = str(result).split(".")
        # cast float to int
        if dec == "0":
            result = int(integer)

    return result


def pt_number_parse(words, i):
    def pt_cte(i, s):
        if i < len(words) and s == words[i]:
            return s, i + 1
        return None

    def pt_number_word(i, mi, ma):
        if i < len(words):
            v = pt_numbers.get(words[i])
            if v and v >= mi and v <= ma:
                return v, i + 1
        return None

    def pt_number_1_99(i):
        r1 = pt_number_word(i, 1, 29)
        if r1:
            return r1

        r1 = pt_number_word(i, 30, 90)
        if r1:
            v1, i1 = r1
            r2 = pt_cte(i1, "e")
            if r2:
                i2 = r2[1]
                r3 = pt_number_word(i2, 1, 9)
                if r3:
                    v3, i3 = r3
                    return v1 + v3, i3
            return r1
        return None

    def pt_number_1_999(i):
        # [2-9]cientos [1-99]?
        r1 = pt_number_word(i, 100, 900)
        if r1:
            v1, i1 = r1
            r2 = pt_number_1_99(i1)
            if r2:
                v2, i2 = r2
                return v1 + v2, i2
            else:
                return r1

        # [1-99]
        r1 = pt_number_1_99(i)
        if r1:
            return r1

        return None

    def pt_number(i):
        # check for cero
        r1 = pt_number_word(i, 0, 0)
        if r1:
            return r1

        # check for [1-999] (mil [0-999])?
        r1 = pt_number_1_999(i)
        if r1:
            v1, i1 = r1
            r2 = pt_cte(i1, "mil")
            if r2:
                i2 = r2[1]
                r3 = pt_number_1_999(i2)
                if r3:
                    v3, i3 = r3
                    return v1 * 1000 + v3, i3
                else:
                    return v1 * 1000, i2
            else:
                return r1
        return None

    return pt_number(i)


def normalize_pt(text, remove_articles):
    """ PT string normalization """

    words = text.split()  # this also removed extra spaces
    normalized = ""
    # Contractions are not common in PT

    # Convert numbers into digits, e.g. "dois" -> "2"
    normalized = ""
    i = 0
    while i < len(words):
        word = words[i]
        # remove articles
        if remove_articles and word in pt_articles:
            i += 1
            continue

        # Convert numbers into digits
        r = pt_number_parse(words, i)
        if r:
            v, i = r
            normalized += " " + str(v)
            continue

        # NOTE temporary , handle some numbers above >999
        if word in pt_numbers:
            word = str(pt_numbers[word])
        # end temporary

        normalized += " " + word
        i += 1
    # some articles in pt-pt can not be removed, but many words can
    # this is experimental and some meaning may be lost
    # maybe agressive should default to False
    # only usage will tell, as a native speaker this seems reasonable
    return pt_pruning(normalized[1:], agressive=remove_articles)


def extract_datetime_pt(input_str, currentDate):
    def clean_string(s):
        # cleans the input string of unneeded punctuation and capitalization
        # among other things
        symbols = [".", ",", ";", "?", "!", u"º", u"ª"]
        noise_words = ["o", "os", "a", "as", "do", "da", "dos", "das", "de",
                       "ao", "aos"]

        for word in symbols:
            s = s.replace(word, "")
        for word in noise_words:
            s = s.replace(" " + word + " ", " ")
        s = s.lower().replace(
            u"á",
            "a").replace(
            u"ç",
            "c").replace(
            u"à",
            "a").replace(
            u"ã",
            "a").replace(
            u"é",
            "e").replace(
            u"è",
            "e").replace(
            u"ê",
            "e").replace(
            u"ó",
            "o").replace(
            u"ò",
            "o").replace(
            "-",
            " ").replace(
            "_",
            "")
        # handle synonims and equivalents, "tomorrow early = tomorrow morning
        synonims = {"manha": ["manhazinha", "cedo", "cedinho"],
                    "tarde": ["tardinha", "tarde"],
                    "noite": ["noitinha", "anoitecer"],
                    "todos": ["ao", "aos"],
                    "em": ["do", "da", "dos", "das", "de"]}
        for syn in synonims:
            for word in synonims[syn]:
                s = s.replace(" " + word + " ", " " + syn + " ")
        # relevant plurals, cant just extract all s in pt
        wordlist = ["manhas", "noites", "tardes", "dias", "semanas", "anos",
                    "minutos", "segundos", "nas", "nos", "proximas",
                    "seguintes", "horas"]
        for _, word in enumerate(wordlist):
            s = s.replace(word, word.rstrip('s'))
        s = s.replace("meses", "mes").replace("anteriores", "anterior")
        return s

    def date_found():
        return found or \
               (
                       datestr != "" or timeStr != "" or
                       yearOffset != 0 or monthOffset != 0 or
                       dayOffset is True or hrOffset != 0 or
                       hrAbs != 0 or minOffset != 0 or
                       minAbs != 0 or secOffset != 0
               )

    if input_str == "" or not currentDate:
        return None

    found = False
    daySpecified = False
    dayOffset = False
    monthOffset = 0
    yearOffset = 0
    dateNow = currentDate
    today = dateNow.strftime("%w")
    currentYear = dateNow.strftime("%Y")
    fromFlag = False
    datestr = ""
    hasYear = False
    timeQualifier = ""

    words = clean_string(input_str).split(" ")
    timeQualifiersList = ['manha', 'tarde', 'noite']
    time_indicators = ["em", "as", "nas", "pelas", "volta", "depois", "estas",
                       "no", "dia", "hora"]
    days = ['segunda', 'terca', 'quarta',
            'quinta', 'sexta', 'sabado', 'domingo']
    months = ['janeiro', 'febreiro', 'marco', 'abril', 'maio', 'junho',
              'julho', 'agosto', 'setembro', 'outubro', 'novembro',
              'dezembro']
    monthsShort = ['jan', 'feb', 'mar', 'abr', 'mai', 'jun', 'jul', 'ag',
                   'set', 'out', 'nov', 'dec']
    nexts = ["proximo", "proxima"]
    suffix_nexts = ["seguinte", "subsequente", "seguir"]
    lasts = ["ultimo", "ultima"]
    suffix_lasts = ["passada", "passado", "anterior", "antes"]
    nxts = ["depois", "seguir", "seguida", "seguinte", "proxima", "proximo"]
    prevs = ["antes", "ante", "previa", "previamente", "anterior"]
    froms = ["partir", "em", "para", "na", "no", "daqui", "seguir",
             "depois", "por", "proxima", "proximo", "da", "do", "de"]
    thises = ["este", "esta", "deste", "desta", "neste", "nesta", "nesse",
              "nessa"]
    froms += thises
    lists = nxts + prevs + froms + time_indicators
    for idx, word in enumerate(words):
        if word == "":
            continue
        wordPrevPrev = words[idx - 2] if idx > 1 else ""
        wordPrev = words[idx - 1] if idx > 0 else ""
        wordNext = words[idx + 1] if idx + 1 < len(words) else ""
        wordNextNext = words[idx + 2] if idx + 2 < len(words) else ""
        wordNextNextNext = words[idx + 3] if idx + 3 < len(words) else ""

        start = idx
        used = 0
        # save timequalifier for later
        if word in timeQualifiersList:
            timeQualifier = word

        # parse today, tomorrow, yesterday
        elif word == "hoje" and not fromFlag:
            dayOffset = 0
            used += 1
        elif word == "amanha" and not fromFlag:
            dayOffset = 1
            used += 1
        elif word == "ontem" and not fromFlag:
            dayOffset -= 1
            used += 1
        # "before yesterday" and "before before yesterday"
        elif (word == "anteontem" or
              (word == "ante" and wordNext == "ontem")) and not fromFlag:
            dayOffset -= 2
            used += 1
            if wordNext == "ontem":
                used += 1
        elif word == "ante" and wordNext == "ante" and wordNextNext == \
                "ontem" and not fromFlag:
            dayOffset -= 3
            used += 3
        elif word == "anteanteontem" and not fromFlag:
            dayOffset -= 3
            used += 1
        # day after tomorrow
        elif word == "depois" and wordNext == "amanha" and not fromFlag:
            dayOffset += 2
            used = 2
        # day before yesterday
        elif word == "antes" and wordNext == "ontem" and not fromFlag:
            dayOffset -= 2
            used = 2
        # parse 5 days, 10 weeks, last week, next week, week after
        elif word == "dia":
            if wordNext == "depois" or wordNext == "antes":
                used += 1
                if wordPrev and wordPrev[0].isdigit():
                    dayOffset += int(wordPrev)
                    start -= 1
                    used += 1
            elif (wordPrev and wordPrev[0].isdigit() and
                  wordNext not in months and
                  wordNext not in monthsShort):
                dayOffset += int(wordPrev)
                start -= 1
                used += 2
            elif wordNext and wordNext[0].isdigit() and wordNextNext not in \
                    months and wordNextNext not in monthsShort:
                dayOffset += int(wordNext)
                start -= 1
                used += 2

        elif word == "semana" and not fromFlag:
            if wordPrev[0].isdigit():
                dayOffset += int(wordPrev) * 7
                start -= 1
                used = 2
            for w in nexts:
                if wordPrev == w:
                    dayOffset = 7
                    start -= 1
                    used = 2
            for w in lasts:
                if wordPrev == w:
                    dayOffset = -7
                    start -= 1
                    used = 2
            for w in suffix_nexts:
                if wordNext == w:
                    dayOffset = 7
                    start -= 1
                    used = 2
            for w in suffix_lasts:
                if wordNext == w:
                    dayOffset = -7
                    start -= 1
                    used = 2
        # parse 10 months, next month, last month
        elif word == "mes" and not fromFlag:
            if wordPrev[0].isdigit():
                monthOffset = int(wordPrev)
                start -= 1
                used = 2
            for w in nexts:
                if wordPrev == w:
                    monthOffset = 7
                    start -= 1
                    used = 2
            for w in lasts:
                if wordPrev == w:
                    monthOffset = -7
                    start -= 1
                    used = 2
            for w in suffix_nexts:
                if wordNext == w:
                    monthOffset = 7
                    start -= 1
                    used = 2
            for w in suffix_lasts:
                if wordNext == w:
                    monthOffset = -7
                    start -= 1
                    used = 2
        # parse 5 years, next year, last year
        elif word == "ano" and not fromFlag:
            if wordPrev[0].isdigit():
                yearOffset = int(wordPrev)
                start -= 1
                used = 2
            for w in nexts:
                if wordPrev == w:
                    yearOffset = 7
                    start -= 1
                    used = 2
            for w in lasts:
                if wordPrev == w:
                    yearOffset = -7
                    start -= 1
                    used = 2
            for w in suffix_nexts:
                if wordNext == w:
                    yearOffset = 7
                    start -= 1
                    used = 2
            for w in suffix_lasts:
                if wordNext == w:
                    yearOffset = -7
                    start -= 1
                    used = 2
        # parse Monday, Tuesday, etc., and next Monday,
        # last Tuesday, etc.
        elif word in days and not fromFlag:

            d = days.index(word)
            dayOffset = (d + 1) - int(today)
            used = 1
            if dayOffset < 0:
                dayOffset += 7
            for w in nexts:
                if wordPrev == w:
                    dayOffset += 7
                    used += 1
                    start -= 1
            for w in lasts:
                if wordPrev == w:
                    dayOffset -= 7
                    used += 1
                    start -= 1
            for w in suffix_nexts:
                if wordNext == w:
                    dayOffset += 7
                    used += 1
                    start -= 1
            for w in suffix_lasts:
                if wordNext == w:
                    dayOffset -= 7
                    used += 1
                    start -= 1
            if wordNext == "feira":
                used += 1
        # parse 15 of July, June 20th, Feb 18, 19 of February
        elif word in months or word in monthsShort:
            try:
                m = months.index(word)
            except ValueError:
                m = monthsShort.index(word)
            used += 1
            datestr = months[m]
            if wordPrev and wordPrev[0].isdigit():
                # 13 maio
                datestr += " " + wordPrev
                start -= 1
                used += 1
                if wordNext and wordNext[0].isdigit():
                    datestr += " " + wordNext
                    used += 1
                    hasYear = True
                else:
                    hasYear = False

            elif wordNext and wordNext[0].isdigit():
                # maio 13
                datestr += " " + wordNext
                used += 1
                if wordNextNext and wordNextNext[0].isdigit():
                    datestr += " " + wordNextNext
                    used += 1
                    hasYear = True
                else:
                    hasYear = False

            elif wordPrevPrev and wordPrevPrev[0].isdigit():
                # 13 dia maio
                datestr += " " + wordPrevPrev

                start -= 2
                used += 2
                if wordNext and word[0].isdigit():
                    datestr += " " + wordNext
                    used += 1
                    hasYear = True
                else:
                    hasYear = False

            elif wordNextNext and wordNextNext[0].isdigit():
                # maio dia 13
                datestr += " " + wordNextNext
                used += 2
                if wordNextNextNext and wordNextNextNext[0].isdigit():
                    datestr += " " + wordNextNextNext
                    used += 1
                    hasYear = True
                else:
                    hasYear = False

            if datestr in months:
                datestr = ""

        # parse 5 days from tomorrow, 10 weeks from next thursday,
        # 2 months from July
        validFollowups = days + months + monthsShort
        validFollowups.append("hoje")
        validFollowups.append("amanha")
        validFollowups.append("ontem")
        validFollowups.append("anteontem")
        validFollowups.append("agora")
        validFollowups.append("ja")
        validFollowups.append("ante")

        # TODO debug word "depois" that one is failing for some reason
        if word in froms and wordNext in validFollowups:

            if not (wordNext == "amanha" and wordNext == "ontem") and not (
                    word == "depois" or word == "antes" or word == "em"):
                used = 2
                fromFlag = True
            if wordNext == "amanha" and word != "depois":
                dayOffset += 1
            elif wordNext == "ontem":
                dayOffset -= 1
            elif wordNext == "anteontem":
                dayOffset -= 2
            elif wordNext == "ante" and wordNextNext == "ontem":
                dayOffset -= 2
            elif (wordNext == "ante" and wordNext == "ante" and
                  wordNextNextNext == "ontem"):
                dayOffset -= 3
            elif wordNext in days:
                d = days.index(wordNext)
                tmpOffset = (d + 1) - int(today)
                used = 2
                if wordNextNext == "feira":
                    used += 1
                if tmpOffset < 0:
                    tmpOffset += 7
                if wordNextNext:
                    if wordNextNext in nxts:
                        tmpOffset += 7
                        used += 1
                    elif wordNextNext in prevs:
                        tmpOffset -= 7
                        used += 1
                dayOffset += tmpOffset
            elif wordNextNext and wordNextNext in days:
                d = days.index(wordNextNext)
                tmpOffset = (d + 1) - int(today)
                used = 3
                if wordNextNextNext:
                    if wordNextNextNext in nxts:
                        tmpOffset += 7
                        used += 1
                    elif wordNextNextNext in prevs:
                        tmpOffset -= 7
                        used += 1
                dayOffset += tmpOffset
                if wordNextNextNext == "feira":
                    used += 1
        if wordNext in months:
            used -= 1
        if used > 0:

            if start - 1 > 0 and words[start - 1] in lists:
                start -= 1
                used += 1

            for i in range(0, used):
                words[i + start] = ""

            if start - 1 >= 0 and words[start - 1] in lists:
                words[start - 1] = ""
            found = True
            daySpecified = True

    # parse time
    timeStr = ""
    hrOffset = 0
    minOffset = 0
    secOffset = 0
    hrAbs = 0
    minAbs = 0
    military = False

    for idx, word in enumerate(words):
        if word == "":
            continue

        wordPrevPrev = words[idx - 2] if idx > 1 else ""
        wordPrev = words[idx - 1] if idx > 0 else ""
        wordNext = words[idx + 1] if idx + 1 < len(words) else ""
        wordNextNext = words[idx + 2] if idx + 2 < len(words) else ""
        wordNextNextNext = words[idx + 3] if idx + 3 < len(words) else ""
        # parse noon, midnight, morning, afternoon, evening
        used = 0
        if word == "meio" and wordNext == "dia":
            hrAbs = 12
            used += 2
        elif word == "meia" and wordNext == "noite":
            hrAbs = 0
            used += 2
        elif word == "manha":
            if hrAbs == 0:
                hrAbs = 8
            used += 1
        elif word == "tarde":
            if hrAbs == 0:
                hrAbs = 15
            used += 1
        elif word == "meio" and wordNext == "tarde":
            if hrAbs == 0:
                hrAbs = 17
            used += 2
        elif word == "meio" and wordNext == "manha":
            if hrAbs == 0:
                hrAbs = 10
            used += 2
        elif word == "fim" and wordNext == "tarde":
            if hrAbs == 0:
                hrAbs = 19
            used += 2
        elif word == "fim" and wordNext == "manha":
            if hrAbs == 0:
                hrAbs = 11
            used += 2
        elif word == "tantas" and wordNext == "manha":
            if hrAbs == 0:
                hrAbs = 4
            used += 2
        elif word == "noite":
            if hrAbs == 0:
                hrAbs = 22
            used += 1
        # parse half an hour, quarter hour
        elif word == "hora" and \
                (wordPrev in time_indicators or wordPrevPrev in
                 time_indicators):
            if wordPrev == "meia":
                minOffset = 30
            elif wordPrev == "quarto":
                minOffset = 15
            elif wordPrevPrev == "quarto":
                minOffset = 15
                if idx > 2 and words[idx - 3] in time_indicators:
                    words[idx - 3] = ""
                words[idx - 2] = ""
            else:
                hrOffset = 1
            if wordPrevPrev in time_indicators:
                words[idx - 2] = ""
            words[idx - 1] = ""
            used += 1
            hrAbs = -1
            minAbs = -1
        # parse 5:00 am, 12:00 p.m., etc
        elif word[0].isdigit():
            isTime = True
            strHH = ""
            strMM = ""
            remainder = ""
            if ':' in word:
                # parse colons
                # "3:00 in the morning"
                stage = 0
                length = len(word)
                for i in range(length):
                    if stage == 0:
                        if word[i].isdigit():
                            strHH += word[i]
                        elif word[i] == ":":
                            stage = 1
                        else:
                            stage = 2
                            i -= 1
                    elif stage == 1:
                        if word[i].isdigit():
                            strMM += word[i]
                        else:
                            stage = 2
                            i -= 1
                    elif stage == 2:
                        remainder = word[i:].replace(".", "")
                        break
                if remainder == "":
                    nextWord = wordNext.replace(".", "")
                    if nextWord == "am" or nextWord == "pm":
                        remainder = nextWord
                        used += 1
                    elif wordNext == "manha":
                        remainder = "am"
                        used += 1
                    elif wordNext == "tarde":
                        remainder = "pm"
                        used += 1
                    elif wordNext == "noite":
                        if 0 < int(word[0]) < 6:
                            remainder = "am"
                        else:
                            remainder = "pm"
                        used += 1
                    elif wordNext in thises and wordNextNext == "manha":
                        remainder = "am"
                        used = 2
                    elif wordNext in thises and wordNextNext == "tarde":
                        remainder = "pm"
                        used = 2
                    elif wordNext in thises and wordNextNext == "noite":
                        remainder = "pm"
                        used = 2
                    else:
                        if timeQualifier != "":
                            military = True
                            if strHH <= 12 and \
                                    (timeQualifier == "manha" or
                                     timeQualifier == "tarde"):
                                strHH += 12

            else:
                # try to parse # s without colons
                # 5 hours, 10 minutes etc.
                length = len(word)
                strNum = ""
                remainder = ""
                for i in range(length):
                    if word[i].isdigit():
                        strNum += word[i]
                    else:
                        remainder += word[i]

                if remainder == "":
                    remainder = wordNext.replace(".", "").lstrip().rstrip()

                if (
                        remainder == "pm" or
                        wordNext == "pm" or
                        remainder == "p.m." or
                        wordNext == "p.m."):
                    strHH = strNum
                    remainder = "pm"
                    used = 1
                elif (
                        remainder == "am" or
                        wordNext == "am" or
                        remainder == "a.m." or
                        wordNext == "a.m."):
                    strHH = strNum
                    remainder = "am"
                    used = 1
                else:
                    if (wordNext == "pm" or
                            wordNext == "p.m." or
                            wordNext == "tarde"):
                        strHH = strNum
                        remainder = "pm"
                        used = 1
                    elif (wordNext == "am" or
                          wordNext == "a.m." or
                          wordNext == "manha"):
                        strHH = strNum
                        remainder = "am"
                        used = 1
                    elif (int(word) > 100 and
                          (
                                  wordPrev == "o" or
                                  wordPrev == "oh" or
                                  wordPrev == "zero"
                          )):
                        # 0800 hours (pronounced oh-eight-hundred)
                        strHH = int(word) / 100
                        strMM = int(word) - strHH * 100
                        military = True
                        if wordNext == "hora":
                            used += 1
                    elif (
                            wordNext == "hora" and
                            word[0] != '0' and
                            (
                                    int(word) < 100 and
                                    int(word) > 2400
                            )):
                        # ignores military time
                        # "in 3 hours"
                        hrOffset = int(word)
                        used = 2
                        isTime = False
                        hrAbs = -1
                        minAbs = -1

                    elif wordNext == "minuto":
                        # "in 10 minutes"
                        minOffset = int(word)
                        used = 2
                        isTime = False
                        hrAbs = -1
                        minAbs = -1
                    elif wordNext == "segundo":
                        # in 5 seconds
                        secOffset = int(word)
                        used = 2
                        isTime = False
                        hrAbs = -1
                        minAbs = -1
                    elif int(word) > 100:
                        strHH = int(word) / 100
                        strMM = int(word) - strHH * 100
                        military = True
                        if wordNext == "hora":
                            used += 1

                    elif wordNext == "" or (
                            wordNext == "em" and wordNextNext == "ponto"):
                        strHH = word
                        strMM = 00
                        if wordNext == "em" and wordNextNext == "ponto":
                            used += 2
                            if wordNextNextNext == "tarde":
                                remainder = "pm"
                                used += 1
                            elif wordNextNextNext == "manha":
                                remainder = "am"
                                used += 1
                            elif wordNextNextNext == "noite":
                                if 0 > int(strHH) > 6:
                                    remainder = "am"
                                else:
                                    remainder = "pm"
                                used += 1

                    elif wordNext[0].isdigit():
                        strHH = word
                        strMM = wordNext
                        military = True
                        used += 1
                        if wordNextNext == "hora":
                            used += 1
                    else:
                        isTime = False

            strHH = int(strHH) if strHH else 0
            strMM = int(strMM) if strMM else 0
            strHH = strHH + 12 if (remainder == "pm" and
                                   0 < strHH < 12) else strHH
            strHH = strHH - 12 if (remainder == "am" and
                                   0 < strHH >= 12) else strHH
            if strHH > 24 or strMM > 59:
                isTime = False
                used = 0
            if isTime:
                hrAbs = strHH * 1
                minAbs = strMM * 1
                used += 1

        if used > 0:
            # removed parsed words from the sentence
            for i in range(used):
                words[idx + i] = ""

            if wordPrev == "em" or wordPrev == "ponto":
                words[words.index(wordPrev)] = ""

            if idx > 0 and wordPrev in time_indicators:
                words[idx - 1] = ""
            if idx > 1 and wordPrevPrev in time_indicators:
                words[idx - 2] = ""

            idx += used - 1
            found = True

    # check that we found a date
    if not date_found:
        return None

    if dayOffset is False:
        dayOffset = 0

    # perform date manipulation

    extractedDate = dateNow
    extractedDate = extractedDate.replace(microsecond=0,
                                          second=0,
                                          minute=0,
                                          hour=0)
    if datestr != "":
        en_months = ['january', 'february', 'march', 'april', 'may', 'june',
                     'july', 'august', 'september', 'october', 'november',
                     'december']
        en_monthsShort = ['jan', 'feb', 'mar', 'apr', 'may', 'june', 'july',
                          'aug',
                          'sept', 'oct', 'nov', 'dec']
        for idx, en_month in enumerate(en_months):
            datestr = datestr.replace(months[idx], en_month)
        for idx, en_month in enumerate(en_monthsShort):
            datestr = datestr.replace(monthsShort[idx], en_month)

        temp = datetime.strptime(datestr, "%B %d")
        if not hasYear:
            temp = temp.replace(year=extractedDate.year)
            if extractedDate < temp:
                extractedDate = extractedDate.replace(year=int(currentYear),
                                                      month=int(
                                                          temp.strftime(
                                                              "%m")),
                                                      day=int(temp.strftime(
                                                          "%d")))
            else:
                extractedDate = extractedDate.replace(
                    year=int(currentYear) + 1,
                    month=int(temp.strftime("%m")),
                    day=int(temp.strftime("%d")))
        else:
            extractedDate = extractedDate.replace(
                year=int(temp.strftime("%Y")),
                month=int(temp.strftime("%m")),
                day=int(temp.strftime("%d")))

    if timeStr != "":
        temp = datetime(timeStr)
        extractedDate = extractedDate.replace(hour=temp.strftime("%H"),
                                              minute=temp.strftime("%M"),
                                              second=temp.strftime("%S"))

    if yearOffset != 0:
        extractedDate = extractedDate + relativedelta(years=yearOffset)
    if monthOffset != 0:
        extractedDate = extractedDate + relativedelta(months=monthOffset)
    if dayOffset != 0:
        extractedDate = extractedDate + relativedelta(days=dayOffset)
    if hrAbs != -1 and minAbs != -1:

        extractedDate = extractedDate + relativedelta(hours=hrAbs,
                                                      minutes=minAbs)
        if (hrAbs != 0 or minAbs != 0) and datestr == "":
            if not daySpecified and dateNow > extractedDate:
                extractedDate = extractedDate + relativedelta(days=1)
    if hrOffset != 0:
        extractedDate = extractedDate + relativedelta(hours=hrOffset)
    if minOffset != 0:
        extractedDate = extractedDate + relativedelta(minutes=minOffset)
    if secOffset != 0:
        extractedDate = extractedDate + relativedelta(seconds=secOffset)

    resultStr = " ".join(words)
    resultStr = ' '.join(resultStr.split())
    resultStr = pt_pruning(resultStr)
    return [extractedDate, resultStr]


def pt_pruning(text, symbols=True, accents=True, agressive=True):
    # agressive pt word pruning
    words = ["a", "o", "os", "as", "de", "dos", "das",
             "lhe", "lhes", "me", "e", "no", "nas", "na", "nos", "em", "para",
             "este",
             "esta", "deste", "desta", "neste", "nesta", "nesse",
             "nessa", "foi", "que"]
    if symbols:
        symbols = [".", ",", ";", ":", "!", "?", u"ï¿½", u"ï¿½"]
        for symbol in symbols:
            text = text.replace(symbol, "")
        text = text.replace("-", " ").replace("_", " ")
    if accents:
        accents = {"a": [u"á", u"à", u"ã", u"â"],
                   "e": [u"ê", u"è", u"é"],
                   "i": [u"í", u"ì"],
                   "o": [u"ò", u"ó"],
                   "u": [u"ú", u"ù"],
                   "c": [u"ç"]}
        for char in accents:
            for acc in accents[char]:
                text = text.replace(acc, char)
    if agressive:
        text_words = text.split(" ")
        for idx, word in enumerate(text_words):
            if word in words:
                text_words[idx] = ""
        text = " ".join(text_words)
        text = ' '.join(text.split())
    return text


def get_gender_pt(word, raw_string=""):
    word = word.rstrip("s")
    gender = False
    words = raw_string.split(" ")
    for idx, w in enumerate(words):
        if w == word and idx != 0:
            previous = words[idx - 1]
            gender = get_gender_pt(previous)
            break
    if not gender:
        if word[-1] == "a":
            gender = "f"
        if word[-1] == "o" or word[-1] == "e":
            gender = "m"
    return gender



import random
from tnalagmes.data.oregon_trail_data import TERMINOLOGY, RANDOM_EVENTS, GAME_EVENTS
from tnalagmes.engines import TNaLaGmesEngine, Event
from tnalagmes.engines.oregon import SimpleInventory, TurnState, Calendar
from datetime import date


class Oregon75Engine(TNaLaGmesEngine):
    """ so called because logic is ported from 1975 basic version of oregon trail"""
    # https://www.filfre.net/misc/oregon1975.bas

    TERMINOLOGY = TERMINOLOGY
    DATA = GAME_EVENTS
    RANDOM_EVENTS = RANDOM_EVENTS
    name = "Oregon75Engine"

    def __init__(self, from_json=False):
        TNaLaGmesEngine.__init__(self, from_json)
        self.turn = TurnState()
        self.inventory = SimpleInventory()

        if from_json:
            self.import_game_data()

    @property
    def talking_objects(self):
        return [self.inventory, self.turn, self.calendar, self.tracker]

    @property
    def chance_encounter_seed(self):
        # The convoluted original logic for a rider attack: RND(0)*10>((M/100-4)^2+72)/((M/100-4)^2+12)-1
        return (random.random() * 10) > ((
                                                 self.tracker.mileage / 100.0 - 4) ** 2 + 72) / (
                       (
                               self.tracker.mileage / 100.0 - 4) ** 2 + 12) - 1

    def register_default_intents(self):
        # engine interface
        self.register_intent("save", ["save {file}", "save"], self.handle_save)
        self.register_intent("load", ["load {file}", "load"], self.handle_load)
        self.register_intent("export", ["export {file}"], self.handle_export)
        self.register_intent("import", ["import {file}"], self.handle_import)
        self.register_intent("quit", ["quit", "exit", "shutdown", "abort"], self.handle_quit)
        return
        self.register_intent("inventory", ["inventory", "backpack", "case", "briefcase",
                                           "pockets", "slots", "stash", "inv"], self.handle_inventory)
        self.register_intent("health", ["status", "turn data", "health"], self.handle_quit)
        self.register_intent("distance", ["distance"], self.handle_quit)
        self.register_intent("currency", ["currency", "coins", "coin", "cash", "money", "how much", "value"],
                             self.handle_quit)
        self.register_intent("supplies", ["count supplies", "how many supplies", "supply count", "supply number"],
                             self.handle_quit)
        self.register_intent("medicine", ["count medicine", "how many medicine", "medicine count", "medicine number"],
                             self.handle_quit)
        self.register_intent("fuel", ["count fuel", "how many fuel", "fuel count", "fuel number"],
                             self.handle_quit)
        self.register_intent("armour", ["count armour", "how many armour", "armour count", "armour number"],
                             self.handle_quit)
        self.register_intent("ammunition", ["count ammunition", "how many ammunition", "armour ammunition",
                                            "ammunition number"],
                             self.handle_quit)
        self.register_intent("date", ["day", "date", "weeks"],
                             self.handle_quit)
        self.register_intent("location", ["location", "place", "room", "look"],
                             self.handle_quit)

    # turn events
    def on_win(self):
        self.output = self.DATA.get("win", {}).get("intro", "")
        # self.output = "AFTER " + str(self.objective.total_distance) + " LONG MILES---HOORAY!!!!!")
        self.calendar.rollback_date(int(self.tracker.last_turn_fraction * self.calendar.days_per_turn))
        self.output = self.calendar.pretty_date
        self.output = self.inventory.pretty_inventory
        self.output = self.DATA.get("win", {}).get("conclusion", "")

    def on_lose(self):
        self.output = self.DATA.get("lose", {}).get("intro", "")
        # Responses to the first two questions are ignored intentionally
        response = True
        for question in self.DATA.get("lose", "").get("yes_no_questions", []):
            response = self.ask_yes_no(question)
        if not response:
            self.output = self.DATA.get("lose", {}).get("error", "")
        self.output = self.DATA.get("lose", {}).get("conclusion", "")

    def on_maintenance(self):
        response = self.ask_numeric(self.DATA.get("maintenance", {}).get("intro", ""), 1, 3)
        food_eaten = 8 + 5 * response
        if self.inventory.supplies.value < food_eaten:
            self.output = self.DATA.get("maintenance", {}).get("error", "")
            return self.on_maintenance()
        self.turn.eating_state = response
        self.inventory.supplies.subtract(food_eaten)

        self.output = self.DATA.get("maintenance", {}).get("conclusion", "")

    def on_damage(self):
        self.output = self.DATA.get("damage", {}).get("intro", "")
        if (100 * random.random()) < (10 + (35 * self.turn.eating_state - 1)):
            self.output = self.DATA.get("damage", {}).get("mild", "")
            self.tracker.subtract_mileage(5)
            self.inventory.medicine.subtract(2)
        elif (100 * random.random()) < (100 - (40 / (4 ** (self.turn.eating_state - 1)))):
            self.output = self.DATA.get("damage", {}).get("high", "")
            self.tracker.subtract_mileage(5)
            self.inventory.medicine.subtract(5)
        if self.inventory.medicine.value < 0:
            if self.turn.injured:
                self.output = self.DATA.get("damage", {}).get("error", "") + "INJURIES"
            else:
                self.output = self.DATA.get("damage", {}).get("error", "") + "PNEUMONIA"
            self.on_game_over()
        return

    def on_chance_encounter(self):
        peaceful = True
        self.output = self.get_entity("enemy") + self.DATA.get("enemy_encounter", {}).get("intro", "")
        if random.random() < 0.8:
            self.output = self.DATA.get("enemy_encounter", {}).get("peaceful_intro", "")
        else:
            self.output = self.DATA.get("enemy_encounter", {}).get("hostile_intro", "")
            peaceful = False

        # riders may randomly switch sides
        if random.random() <= 0.2:
            peaceful = not peaceful

        response = self.ask_numeric(self.DATA.get("enemy_encounter", {}).get("number_questions", "")[0], 1, 4)
        if not peaceful:
            if response == 1:  # run
                self.tracker.add_mileage(20)
                self.inventory.medicine.subtract(15)
                self.inventory.ammunition.subtract(150)
                self.inventory.fuel.subtract(40)
            elif response == 2:  # attack
                response, entry_time = self.ask_with_timeout()
                # Original bullet loss was "B=B-B1*40-80". This produces a gain in ammunitions
                # when response time is less than 2 seconds and small losses when the value is longer (max: 200)
                self.inventory.ammunition.subtract(entry_time * 28.57)
                if entry_time <= 1:
                    self.output = self.DATA.get("enemy_encounter", {}).get("high_score", "")
                elif entry_time <= 4:
                    self.output = self.DATA.get("enemy_encounter", {}).get("low_score", "")
                else:
                    self.output = self.DATA.get("enemy_encounter", {}).get("damage", "")
                    self.turn.injured = True
            elif response == 3:  # continue
                if random.random() <= 0.8:
                    self.inventory.medicine.subtract(15)
                    self.inventory.ammunition.subtract(150)
                else:
                    self.output = self.DATA.get("enemy_encounter", {}).get("enemy_run", "")
                    return
            else:  # circle wagons
                response, entry_time = self.ask_with_timeout()
                self.inventory.ammunition.subtract((entry_time * 30) - 80)
                self.tracker.subtract_mileage(25)
                if entry_time <= 1:
                    self.output = self.DATA.get("enemy_encounter", {}).get("high_score", "")
                elif entry_time <= 4:
                    self.output = self.DATA.get("enemy_encounter", {}).get("low_score", "")
                else:
                    self.output = self.DATA.get("enemy_encounter", {}).get("damage", "")
                    self.turn.injured = True
        else:  # peaceful riders
            if response == 1:  # run
                self.tracker.add_mileage(15)
                self.inventory.fuel.subtract(10)
            elif response == 2:  # attack
                self.tracker.subtract_mileage(5)
                self.inventory.ammunition.subtract(100)
            elif response == 4:  # circle wagons
                self.tracker.subtract_mileage(20)

        if peaceful:
            self.output = self.get_entity("enemy") + self.DATA.get("enemy_encounter", {}).get("peaceful_conclusion", "")
        else:
            self.output = self.get_entity("enemy") + self.DATA.get("enemy_encounter", {}).get("hostile_conclusion", "")
            if self.inventory.ammunition.value < 0:
                self.output = self.DATA.get("enemy_encounter", {}).get("die", "") + self.get_entity("enemy")
                self.on_game_over()

    def on_easy_difficulty(self):
        # medium difficulty
        if (random.random() * 10) <= (
                9 - ((self.tracker.mileage / 100 - 15) ** 2 + 72) / ((self.tracker.mileage / 100 - 15) ** 2 + 12)):
            self.output = self.DATA.get("easy_difficulty", {}).get("intro", "")
            if random.random() <= 0.1:
                self.output = self.DATA.get("easy_difficulty", {}).get("events", "")[0]
                self.tracker.subtract_mileage(60)
            elif random.random() <= 0.11:
                self.output = self.DATA.get("easy_difficulty", {}).get("events", "")[1]
                self.tracker.subtract_mileage(20 + (30 * random.random()))
                self.inventory.medicine.subtract(5)
                self.inventory.ammunition.subtract(200)
            else:
                self.output = self.DATA.get("easy_difficulty", {}).get("events", "")[2]
                self.tracker.subtract_mileage(45 + (random.random() // 0.02))

        # First pass evaluated at 950 miles (reached_mountains)
        if not self.tracker.medium_difficulty:
            self.tracker.medium_difficulty = True

    def on_medium_difficulty(self):
        # First pass evaluated at 950 miles (reached_mountains)
        if random.random() < 0.8:
            self.on_difficulty_damage()
        else:
            self.output = self.DATA.get("medium_difficulty", {}).get("conclusion", "")
            self.tracker.medium_difficulty = False
        if self.tracker.mileage >= 1700 and not self.tracker.hard_difficulty:
            self.tracker.hard_difficulty = True

    def on_hard_difficulty(self):
        # Second pass (blue mountains) at 1700 miles
        if random.random() < 0.7:
            self.on_difficulty_damage()

    def on_explore(self):
        if self.inventory.ammunition.value < 40:
            self.output = self.DATA.get("explore", {}).get("error", "")
            return
        self.tracker.subtract_mileage(45)

        response, entry_time = self.ask_with_timeout()
        # debug logging? print "User typed", response, "after", entry_time, "seconds"

        if entry_time < 1.0:
            self.output = self.DATA.get("explore", {}).get("events", "")[0]
            self.inventory.supplies.add(52 + random.random() * 6)
            self.inventory.ammunition.subtract(10 - random.random() * 4)
        elif (100 * random.random()) < (13 * entry_time):
            self.output = self.DATA.get("explore", {}).get("events", "")[1]
        else:
            self.output = self.DATA.get("explore", {}).get("events", "")[2]
            self.inventory.supplies.add(48 - 2 * entry_time)
            self.inventory.ammunition.subtract(10 - 3 * entry_time)

    def on_shop(self):
        self.output = self.DATA.get("shop", {}).get("intro", "")
        food = self.ask_numeric(self.get_entity("supplies"), 0, self.inventory.currency)
        ammo = self.ask_numeric(self.get_entity("ammunition"), 0, self.inventory.currency)
        clothing = self.ask_numeric(self.get_entity("armour"), 0, self.inventory.currency)
        misc = self.ask_numeric(self.get_entity("medicine"), 0, self.inventory.currency)
        total_spend = food + ammo + clothing + misc
        if self.inventory.currency < total_spend:
            self.output = self.DATA.get("shop", {}).get("error", "")
            return self.on_shop
        self.inventory.spend(total_spend)
        self.inventory.supplies.add(0.66 * food)
        self.inventory.ammunition.add(0.66 * ammo * 50)
        self.inventory.armour.add(0.66 * clothing)
        self.inventory.medicine.add(0.66 * misc)
        self.tracker.subtract_mileage(45)

    def on_heal(self):
        if self.turn.illness or self.turn.injured:
            self.inventory.spend(20)
            if self.inventory.currency < 0:
                self.output = self.DATA.get("heal", {}).get("error", "")
                if self.turn.illness:
                    self.output = self.DATA.get("heal", {}).get("die", "") + self.get_entity("passive_damage")
                elif self.turn.injured:
                    self.output = self.DATA.get("heal", {}).get("die", "") + self.get_entity("attack_damage")
                return False
            else:
                self.output = self.DATA.get("heal", {}).get("conclusion", "")
                self.turn.illness = False
                self.turn.injured = False

        return True

    def on_turn(self):
        if self.tracker.completed:
            return
        self.output = self.calendar.pretty_date
        self.inventory.normalize_negative_values()

        # Resolve health issues from the previous turn
        if not self.on_heal():
            self.on_game_over()
            return

        # Show inventory status and mileage
        self.output = self.inventory.warnings
        self.tracker.print_mileage()

        # Ask for turn options
        self.output = self.inventory.pretty_inventory
        turn_response = self.ask_numeric(self.DATA.get("turn", {}).get("intro", ""), 1, 3)
        if turn_response == 1:
            self.on_shop()
        elif turn_response == 2:
            self.on_explore()

        # Eating
        if self.inventory.supplies.value < 13:
            self.output = self.DATA.get("turn", {}).get("die", "")
            self.on_game_over()
            return

        self.on_maintenance()

        # Advance mileage now, events may subtract from overall
        # progress for each turn
        self.tracker.random_advance(self.inventory.fuel.value)

        # Rider attack
        if self.chance_encounter_seed:
            self.on_chance_encounter()

        # Random per turn events
        self.on_random_event()

        # Mountain events
        self.on_difficulty_modifier()

        # Move to next turn
        self.calendar.advance_date()

        # turns

    def on_difficulty_damage(self):
        self.output = self.DATA.get("difficulty_damage", {}).get("intro", "")
        self.inventory.supplies.subtract(25)
        self.inventory.medicine.subtract(10)
        self.inventory.ammunition.subtract(300)
        self.tracker.subtract_mileage(30 + (40 * random.random()))
        if self.inventory.armour.value < (18 + (2 * random.random())):
            self.on_damage()

    # random events
    def rain(self):
        if self.tracker.difficulty_triggered():
            self.output = self.RANDOM_EVENTS.get("rain", {}).get("intro", "")
            self.inventory.supplies.subtract(10)
            self.inventory.ammunition.subtract(500)
            self.inventory.medicine.subtract(15)
            self.tracker.subtract_mileage((10 * random.random()) + 5)
            self.output = self.RANDOM_EVENTS.get("conclusion", "")

    def storm(self):
        self.output = self.RANDOM_EVENTS.get("storm", {}).get("intro")
        self.tracker.subtract_mileage(5 + (10 * random.random()))
        self.inventory.ammunition.subtract(200)
        self.inventory.medicine.subtract(4 + (3 * random.random()))

    def cold(self):
        self.output = self.RANDOM_EVENTS.get("cold", {}).get("intro", "")
        insufficient_clothing = False
        if self.inventory.armour.value < (22 + (4 * random.random())):
            self.output = self.RANDOM_EVENTS.get("cold", {}).get("error", "")
            insufficient_clothing = True
        self.output = self.RANDOM_EVENTS.get("cold", {}).get("conclusion", "")
        if insufficient_clothing:
            self.on_damage()

    def enemy_attack(self):
        self.output = self.RANDOM_EVENTS.get("attack", {}).get("intro", "")
        response, entry_time = self.ask_with_timeout()
        self.inventory.ammunition.subtract(20 * entry_time)
        if self.inventory.ammunition.value <= 0:
            self.output = self.RANDOM_EVENTS.get("attack", {}).get("error", "")
            self.inventory.spend(self.inventory.currency * 0.66)
        elif entry_time <= 1:
            self.output = self.RANDOM_EVENTS.get("attack", {}).get("conclusion", "")
        else:
            self.output = self.RANDOM_EVENTS.get("attack", {}).get("damage", "")
            self.turn.injured = True
            self.inventory.medicine.subtract(5)
            self.inventory.fuel.subtract(20)

    def get_poisoned(self):
        self.output = self.RANDOM_EVENTS.get("poison", {}).get("intro", "")
        self.inventory.ammunition.subtract(10)
        self.inventory.medicine.subtract(5)
        if self.inventory.medicine.value <= 0:
            # TODO use data
            self.output = "YOU DIE OF SNAKEBITE SINCE YOU HAVE NO MEDICINE"
        self.on_game_over()
        self.output = self.RANDOM_EVENTS.get("poison", {}).get("conclusion", "")

    def animal_attack(self):
        self.output = self.RANDOM_EVENTS.get("animal", {}).get("intro", "")
        response, entry_time = self.ask_with_timeout()
        if self.inventory.ammunition.value < 40:
            self.output = self.RANDOM_EVENTS.get("animal", {}).get("damage", "")
        self.turn.injured = True
        if entry_time <= 2:
            self.output = self.RANDOM_EVENTS.get("animal", {}).get("conclusion", "")
        else:
            self.output = self.RANDOM_EVENTS.get("animal", {}).get("error", "")
        self.inventory.ammunition.subtract(20 * entry_time)
        self.inventory.armour.subtract(4 * entry_time)
        self.inventory.supplies.subtract(8 * entry_time)

    def shelter_damage(self):
        self.output = self.RANDOM_EVENTS.get("shelter_damage", {}).get("intro", "")
        self.tracker.subtract_mileage(15 + 5 * random.random())
        self.inventory.medicine.subtract(8)
        self.output = self.RANDOM_EVENTS.get("shelter_damage", {}).get("conclusion", "")

    def vehicle_damage(self):
        self.output = self.RANDOM_EVENTS.get("vehicle_damage", {}).get("intro", "")
        self.tracker.subtract_mileage(25)
        self.inventory.fuel.subtract(20)
        self.output = self.RANDOM_EVENTS.get("vehicle_damage", {}).get("conclusion", "")

    def fuel_damage(self):
        self.output = self.RANDOM_EVENTS.get("fuel_damage", {}).get("intro", "")
        self.tracker.subtract_mileage(17)
        self.output = self.RANDOM_EVENTS.get("fuel_damage", {}).get("conclusion", "")

    def lose_companion(self):
        self.output = self.RANDOM_EVENTS.get("companion_lose", {}).get("intro", "")
        self.tracker.subtract_mileage(10)
        self.output = self.RANDOM_EVENTS.get("companion_lose", {}).get("conclusion", "")

    def supply_damage(self):
        self.output = self.RANDOM_EVENTS.get("supply_damage", {}).get("intro", "")
        self.tracker.subtract_mileage((10 * random.random()) + 2)
        self.output = self.RANDOM_EVENTS.get("supply_damage", {}).get("conclusion", "")

    def shelter_fire(self):
        self.output = self.RANDOM_EVENTS.get("shelter_fire", {}).get("intro", "")
        self.inventory.supplies.subtract(40)
        self.inventory.ammunition.subtract(400)
        self.inventory.medicine.subtract((random.random() * 8) + 3)
        self.tracker.subtract_mileage(15)
        self.output = self.RANDOM_EVENTS.get("shelter_fire", {}).get("conclusion", "")

    def heavy_fog(self):
        self.output = self.RANDOM_EVENTS.get("fog", {}).get("intro", "")
        self.tracker.subtract_mileage(10 + (5 * random.random()))
        self.output = self.RANDOM_EVENTS.get("fog", {}).get("conclusion", "")

    def bad_terrain(self):
        self.output = self.RANDOM_EVENTS.get("bad_terrain", {}).get("intro", "")
        self.inventory.supplies.subtract(30)
        self.inventory.armour.subtract(20)
        self.tracker.subtract_mileage(20 + (20 * random.random()))
        self.output = self.RANDOM_EVENTS.get("bad_terrain", {}).get("conclusion", "")

    def find_supplies(self):
        self.output = self.RANDOM_EVENTS.get("find_supplies", {}).get("intro", "")
        self.inventory.supplies.add(14)
        self.output = self.RANDOM_EVENTS.get("find_supplies", {}).get("conclusion", "")

    def companion_injury(self):
        self.output = self.RANDOM_EVENTS.get("companion", {}).get("intro", "")
        self.tracker.subtract_mileage(5 + 4 * random.random())
        self.inventory.medicine.subtract(2 + 3 * random.random())
        self.output = self.RANDOM_EVENTS.get("companion", {}).get("conclusion", "")

    def illness(self):
        self.output = self.RANDOM_EVENTS.get("illness", {}).get("intro", "")
        if self.turn.eating_poorly():
            self.on_damage()
        elif self.turn.eating_moderately() and random.random() > 0.25:
            self.on_damage()
        elif self.turn.eating_well() and random.random() < 0.5:
            self.on_damage()
        self.output = self.RANDOM_EVENTS.get("illness", {}).get("conclusion", "")

    # engine
    def handle_yes(self, intent):
        self.submit_command("yes")
        return ""

    def handle_no(self, intent):
        self.submit_command("no")
        return ""

    def manual_fix_parse(self, text):
        # replace vars
        text = text.replace("{inv.money}",
                            str(self.inventory.currency))
        text = text.replace("{inv.fuel}",
                            str(self.inventory.fuel.value))
        text = text.replace("{inv.supplies}",
                            str(self.inventory.supplies.value))
        text = text.replace("{inv.medicine}",
                            str(self.inventory.medicine.value))
        text = text.replace("{inv.ammunition}",
                            str(self.inventory.ammunition.value))
        text = text.replace("{inv.armour}",
                            str(self.inventory.armour.value))
        text = text.replace("{journey.distance}",
                            str(self.tracker.total_distance))
        return text + "\n"

    def on_start(self):
        self.running = True
        if self.ask_yes_no("Do you need instructions?"):
            self.intro()
        self.inventory = SimpleInventory()
        vehicle_spend = self.ask_numeric(
            self.DATA.get("inventory", {}).get("intro", "") +
            self.TERMINOLOGY["fuel"][0], 200, 300)
        self.inventory.spend(vehicle_spend)
        food_spend = self.ask_numeric(
            self.DATA.get("inventory", {}).get("intro", "") +
            self.TERMINOLOGY["supplies"][0], 0)
        self.inventory.spend(food_spend)
        ammunition_spend = self.ask_numeric(
            self.DATA.get("inventory", {}).get("intro", "") +
            self.TERMINOLOGY["ammunition"][0], 0)
        self.inventory.spend(ammunition_spend)
        clothing_spend = self.ask_numeric(
            self.DATA.get("inventory", {}).get("intro", "") +
            self.TERMINOLOGY["armour"][0], 0)
        self.inventory.spend(clothing_spend)
        misc_spend = self.ask_numeric(
            self.DATA.get("inventory", {}).get("intro", "") +
            self.TERMINOLOGY["medicine"][0], 0)
        self.inventory.spend(misc_spend)

        if self.inventory.currency < 0:
            self.output = self.DATA.get("inventory", {}).get("error", "")
            return self.on_start()

        self.inventory.fuel.value = vehicle_spend
        self.inventory.supplies.value = food_spend
        self.inventory.ammunition.value = 50 * ammunition_spend
        self.inventory.armour.value = clothing_spend
        self.inventory.medicine.value = misc_spend

        self.output = self.DATA.get("inventory", {}).get("conclusion", "")

    def register_events(self):
        """
        load default events or from jsom
        :param from_json:
        :return:
        """
        if self.from_json:
            for event_type in self.RANDOM_EVENTS:
                event = Event()
                event.from_json(self.RANDOM_EVENTS.get(event_type, {}))
                self.register_event(event)
        else:
            self.register_event(self.shelter_damage)
            self.register_event(self.shelter_fire)
            self.register_event(self.lose_companion)
            self.register_event(self.supply_damage)
            self.register_event(self.fuel_damage)
            self.register_event(self.vehicle_damage)
            self.register_event(self.heavy_fog)
            self.register_event(self.bad_terrain)
            self.register_event(self.find_supplies)
            self.register_event(self.companion_injury)
            self.register_event(self.illness)
            self.register_event(self.rain)
            self.register_event(self.animal_attack)
            self.register_event(self.get_poisoned)
            self.register_event(self.enemy_attack)
            self.register_event(self.cold)
            self.register_event(self.storm)


if __name__ == "__main__":
    game = Oregon75Engine()
    game.run()



import pronouncing


def guess_phonemes_en(word):
    word = word.lower()
    basicPronunciations = {'a': ['AE'], 'b': ['B'], 'c': ['K'],
                           'd': ['D'],
                           'e': ['EH'], 'f': ['F'], 'g': ['G'],
                           'h': ['HH'],
                           'i': ['IH'],
                           'j': ['JH'], 'k': ['K'], 'l': ['L'],
                           'm': ['M'],
                           'n': ['N'], 'o': ['OW'], 'p': ['P'],
                           'qu': ['K', 'W'], 'r': ['R'],
                           's': ['S'], 't': ['T'], 'u': ['AH'],
                           'v': ['V'],
                           'w': ['W'], 'x': ['K', 'S'], 'y': ['Y'],
                           'z': ['Z'], 'ch': ['CH'],
                           'sh': ['SH'], 'th': ['TH'], 'dg': ['JH'],
                           'dge': ['JH'], 'psy': ['S', 'AY'],
                           'oi': ['OY'],
                           'ee': ['IY'],
                           'ao': ['AW'], 'ck': ['K'], 'tt': ['T'],
                           'nn': ['N'], 'ai': ['EY'], 'eu': ['Y', 'UW'],
                           'ue': ['UW'],
                           'ie': ['IY'], 'ei': ['IY'], 'ea': ['IY'],
                           'ght': ['T'], 'ph': ['F'], 'gn': ['N'],
                           'kn': ['N'], 'wh': ['W'],
                           'wr': ['R'], 'gg': ['G'], 'ff': ['F'],
                           'oo': ['UW'], 'ua': ['W', 'AO'], 'ng': ['NG'],
                           'bb': ['B'],
                           'tch': ['CH'], 'rr': ['R'], 'dd': ['D'],
                           'cc': ['K', 'S'], 'oe': ['OW'],
                           'igh': ['AY'], 'eigh': ['EY']}
    phones = []

    progress = len(word) - 1
    while progress >= 0:
        if word[0:3] in basicPronunciations.keys():
            for phone in basicPronunciations[word[0:3]]:
                phones.append(phone)
            word = word[3:]
            progress -= 3
        elif word[0:2] in basicPronunciations.keys():
            for phone in basicPronunciations[word[0:2]]:
                phones.append(phone)
            word = word[2:]
            progress -= 2
        elif word[0] in basicPronunciations.keys():
            for phone in basicPronunciations[word[0]]:
                phones.append(phone)
            word = word[1:]
            progress -= 1
        else:
            return None
    return phones


def get_phonemes_en(name):
    name = name.lower()
    phonemes = None
    if " " in name:
        total_phonemes = []
        names = name.split(" ")
        for name in names:
            phon = get_phonemes_en(name)
            if phon is None:
                return None
            total_phonemes.extend(phon)
            total_phonemes.append(" . ")
        if total_phonemes[-1] == " . ":
            total_phonemes = total_phonemes[:-1]
        phonemes = "".join(total_phonemes)
    elif len(pronouncing.phones_for_word(name)):
        phonemes = "".join(pronouncing.phones_for_word(name)[0])
    else:
        guess = guess_phonemes_en(name)
        if guess is not None:
            phonemes = " ".join(guess)

    return phonemes



import random
from tnalagmes.data.template_data import TERMINOLOGY, RANDOM_EVENTS, GAME_EVENTS
from tnalagmes.models.objects import Calendar, InventoryItem, ProgressTracker, Inventory
from tnalagmes import TNaLaGmesConstruct
from tnalagmes.engines import TNaLaGmesEngine


class TurnState(TNaLaGmesConstruct):
    """
    am i injured
    am i eating/poorly/moderately/well
    suffering illness
    """

    def __init__(self):
        TNaLaGmesConstruct.__init__(self, "turn_state")
        self._injured = False
        self._ill = False
        self.bad_luck = False
        self._maintenance_state = 0

    @property
    def injured(self):
        return self._injured

    @injured.setter
    def injured(self, value):
        self._injured = value

    @property
    def illness(self):
        return self._ill

    @illness.setter
    def illness(self, value):
        self._ill = value

    @property
    def eating_state(self):
        return self._maintenance_state

    @eating_state.setter
    def eating_state(self, value):
        if value < 1 or value > 3:
            self._maintenance_state = 0
        else:
            self._maintenance_state = value

    def eating_poorly(self):
        return self._maintenance_state == 1

    def eating_moderately(self):
        return self._maintenance_state == 2

    def eating_well(self):
        return self._maintenance_state == 3


class SimpleInventory(Inventory):
    TERMINOLOGY = TERMINOLOGY

    def __init__(self, start_money=700):
        TNaLaGmesConstruct.__init__(self, "inventory")
        self.start_money = start_money
        self._money = start_money
        self.fuel = InventoryItem()
        self.supplies = InventoryItem()
        self.ammunition = InventoryItem()
        self.armour = InventoryItem()
        self.medicine = InventoryItem()

    @classmethod
    def get_entity(cls, text):
        return random.choice(cls.TERMINOLOGY.get(text, [""]))

    @property
    def currency(self):
        return self._money

    def spend(self, cost):
        self._money -= int(cost)

    @property
    def warnings(self):
        if self.supplies.value < 12:
            return GAME_EVENTS.get("low_supplies", {}).get("intro", "")

    @property
    def pretty_inventory(self):
        text = "{} - {}\n".format(self.get_entity("supplies"), self.supplies.value)
        text += "{} - {}\n".format(self.get_entity("ammunition"), self.ammunition.value)
        text += "{} - {}\n".format(self.get_entity("armour"), self.armour.value)
        text += "{} - {}\n".format(self.get_entity("medicine"), self.medicine.value)
        text += "{} - {}\n".format(self.get_entity("currency"), self.currency)
        return text

    def normalize_negative_values(self):
        if self.fuel.value < 0:
            self.fuel.value = 0
        if self.supplies.value < 0:
            self.supplies.value = 0
        if self.ammunition.value < 0:
            self.ammunition.value = 0
        if self.armour.value < 0:
            self.armour.value = 0
        if self.medicine.value < 0:
            self.medicine.value = 0


class OregonEngine(TNaLaGmesEngine):
    """ so called because logic is based on oregon trail"""
    # TODO, bellow is old code from oregon75
    TERMINOLOGY = TERMINOLOGY
    DATA = GAME_EVENTS
    RANDOM_EVENTS = RANDOM_EVENTS
    name = "OregonEngine"

    def __init__(self, name=None, from_json=False):
        TNaLaGmesEngine.__init__(self, "game_engine", from_json)
        self.turn = TurnState()

    def register_default_intents(self):
        # engine interface
        self.register_intent("save", ["save {file}", "save"], self.handle_save)
        self.register_intent("load", ["load {file}", "load"], self.handle_load)
        self.register_intent("export", ["export {file}"], self.handle_export)
        self.register_intent("import", ["import {file}"], self.handle_import)
        self.register_intent("quit", ["quit", "exit", "shutdown", "abort"], self.handle_quit)
        return



#!/usr/bin/python
import random
from threading import Thread
from tnalagmes import TNaLaGmesConstruct
from tnalagmes.data.template_data import TERMINOLOGY, RANDOM_EVENTS, GAME_EVENTS
from tnalagmes.models.objects import Calendar, Inventory, ProgressTracker
from pprint import pprint
from os.path import expanduser, join, exists
from os import makedirs
import json


class Event(TNaLaGmesConstruct):

    def __init__(self, name="", intro="", conclusion="", handler=None):
        TNaLaGmesConstruct.__init__(self, "event")
        self._name = name
        self._intro = intro
        self._conclusion = conclusion
        self.fields = {}
        self._num_triggers = 0
        self.event_handler = handler

    @property
    def num_triggers(self):
        return self._num_triggers

    @property
    def name(self):
        return self._name

    @property
    def intro(self):
        return self._intro

    @property
    def conclusion(self):
        return self._conclusion

    def add_field(self, name, data=None):
        # TODO set property
        self.fields[name] = data

    def from_json(self, data):
        for k in data:
            if k == "_name":
                self._name = data[k]
            elif k == "_intro":
                self._intro = data[k]
            elif k == "_conclusion":
                self._conclusion = data[k]
            else:
                self.add_field(k, data[k])

    @property
    def data(self):
        event_template = {

            "_name": self._name,
            "_intro": self._intro,
            "_conclusion": self._conclusion
        }
        for field in self.fields:
            event_template[field] = self.fields[field]
        return event_template

    def trigger(self, data=None):
        if self.event_handler is not None:
            try:
                return self.event_handler(data)
            except TypeError:
                return self.event_handler()

        self.output = self.intro
        self.output = self.conclusion
        return self.output

    def bind_handler(self, event_handler=None):
        self.event_handler = event_handler


class TNaLaGmesEngine(TNaLaGmesConstruct):
    TERMINOLOGY = TERMINOLOGY
    DATA = GAME_EVENTS
    RANDOM_EVENTS = RANDOM_EVENTS
    name = "TNaLaGmesEngine"

    def __init__(self, from_json=True):
        TNaLaGmesConstruct.__init__(self, "game_engine", scene=self)
        self.from_json = from_json
        self.random_events = []
        self.calendar = Calendar()
        self.tracker = ProgressTracker()
        self.running = False
        self._thread = Thread(target=self._run)
        self._thread.setDaemon(True)

    def on_random_event(self):
        event = random.choice(self.random_events)
        self.output = event.trigger()

    @classmethod
    def get_entity(cls, text):
        return random.choice(cls.TERMINOLOGY.get(text, [""]))

    def pprint_data(self):
        data = {"random_events": self.RANDOM_EVENTS,
                "terminology": self.TERMINOLOGY,
                "turn_data": self.DATA}
        pprint(data)

    def register_core_intents(self):
        # TODO translate options strings to 1, 2, 3, 4

        # engine interface
        self.register_intent("save", ["save {file}", "save"], self.handle_save)
        self.register_intent("load", ["load {file}", "load"], self.handle_load)
        self.register_intent("export", ["export {file}"], self.handle_export)
        self.register_intent("import", ["import {file}"], self.handle_import)
        self.register_intent("quit", ["quit", "exit", "shutdown", "abort"], self.handle_quit)

    def register_event(self, event_object):
        if isinstance(event_object, Event):
            self.random_events.append(event_object)
        elif isinstance(event_object, dict):
            self.register_from_json(event_object)
        elif isinstance(event_object, str):
            e = Event(intro=event_object)
            self.random_events.append(e)
        elif isinstance(event_object, list):
            for e in event_object:
                self.register_event(e)
        elif isinstance(event_object, bool) or \
                isinstance(event_object, int) or \
                isinstance(event_object, float):
            raise AttributeError("attempted to register invalid event")
        else:
            e = Event()
            e.bind_handler(event_object)
            self.random_events.append(e)

    def register_from_json(self, dictionary, event_handler=None):
        event = Event()
        event.from_json(dictionary)
        if event_handler:
            event.bind_handler(event_handler)
        self.register_event(event)

    def register_events(self):
        """
        load default events or from jsom
        :param from_json:
        :return:
        """
        if self.from_json:
            for event_type in self.RANDOM_EVENTS:
                event = Event()
                event.from_json(self.RANDOM_EVENTS.get(event_type, {}))
                self.register_event(event)

    def intro(self):
        self.output = self.DATA.get("intro",{}).get("intro", "")
        self.output = self.DATA.get("intro", {}).get("conclusion", "")

    def on_turn(self):
        self.output = self.calendar.pretty_date

        self.output = "total progress: " + str(self.tracker.mileage)

        # progress for each turn
        self.tracker.random_advance()

        # Random thing happened
        self.on_chance_encounter()

        # Move to next turn
        self.calendar.advance_date()

    def on_win(self):
        self.calendar.rollback_date(int(self.tracker.last_turn_fraction * self.calendar.days_per_turn))
        self.output = self.calendar.pretty_date
        self.running = False

    def on_lose(self):
        self.running = False

    def on_damage(self):
        pass

    def on_chance_encounter(self):
        pass

    def on_easy_difficulty(self):
        pass

    def on_medium_difficulty(self):
        pass

    def on_hard_difficulty(self):
        pass

    def on_shop(self):
        pass

    def on_game_over(self):
        # Turns have been exhausted or objective has been reached
        if self.tracker.completed:
            self.on_win()
        else:
            self.on_lose()
        self.running = False

    def on_difficulty_modifier(self):
        if not self.tracker.difficulty_triggered():
            return
        self.on_easy_difficulty()
        if self.tracker.medium_difficulty:
            self.on_medium_difficulty()
        if self.tracker.hard_difficulty:
            self.on_hard_difficulty()

    def manual_fix_parse(self, text):
        # any parsing ou might want to reading from game data, ie replace {inv.money}
        return text

    def run(self):
        self._thread.start()
        while self.running:
            if self.waiting_for_user:
                command = input(self.output)
                self.parse_command(command)

    def submit_command(self, text=""):
        self.input = text
        self.waiting_for_user = False

    def _run(self):
        self.running = True
        self.on_start()
        self.register_events()

        while not self.calendar.is_final_turn and not self.tracker.completed:
            self.on_turn()
        self.on_game_over()
        self.running = False

    def save(self, path=None):
        pass

    def load(self, path=None):
        pass

    def quit(self):
        if self.ask_yes_no(self.DATA.get("quit_message",
                                         "really want to quit?")):
            self.running = False
            self.on_game_over()

    def handle_quit(self, intent):
        self.quit()
        return "game exited"

    def handle_save(self, intent):
        self.save(intent.get("file"))
        return "game saved"

    def handle_load(self, intent):
        self.load(intent.get("file"))
        return "game loaded"

    def handle_import(self, intent):
        self.import_game_data(intent["file"])
        return "game data imported"

    def handle_export(self, intent):
        self.export_game_data(intent["file"])
        return "game data exported"

    def ask_world(self, utterance):
        # query self intents first
        intents = self.calc_intents(utterance)
        answer = ""
        for intent in intents:
            answer += self.intent_parser.execute_intent(intent)
        if answer and answer != "?":
            return answer

        # query all game objects
        for obj in self.talking_objects:
            # parse intent
            answer = obj.parse_command(utterance)
            if answer.strip().replace(".", "") != "?":
                return answer

        # submit to internal game engine for handling
        return "?"

    @property
    def talking_objects(self):
        return []

    def parse_command(self, utterance):
        # parse intent
        answer = self.ask_world(utterance)
        if answer.strip().replace(".", "") != "?":
            self.output = answer
        else:
            # fallback
            self.output = self.submit_command(utterance)
        #return self.output

    @classmethod
    def export_game_data(cls, path=None):
        path = path or join(expanduser("~"),
                            "TextSurvivalGames")
        if not exists(path):
            makedirs(path)
        file = join(path, cls.name + ".json")
        data = {"random_events": cls.RANDOM_EVENTS,
                "terminology": cls.TERMINOLOGY,
                "turn_data": cls.DATA}
        with open(file, "w") as f:
            f.write(json.dumps(data))
        pprint(data)

    @classmethod
    def import_game_data(cls, path=None):
        path = path or join(expanduser("~"),
                            "TextSurvivalGames")
        file = join(path, cls.name + ".json")
        if exists(file):
            with open(file, "r") as f:
                data = json.loads(f.read())
            pprint(data)
            cls.DATA = data.get("data", cls.DATA)
            cls.TERMINOLOGY = data.get("data",
                                       cls.TERMINOLOGY)
            cls.RANDOM_EVENTS = data.get("data",
                                         cls.RANDOM_EVENTS)

    def on_start(self):
        if self.ask_yes_no("Do you need instructions?"):
            self.intro()



from random import random, choice, randint
from time import sleep

from tnalagmes.data.template_data import TERMINOLOGY, RANDOM_EVENTS, GAME_EVENTS
from tnalagmes.models import Calendar, InventoryItem, ProgressTracker
from tnalagmes import TNaLaGmesConstruct
from tnalagmes.engines import TNaLaGmesEngine
from os.path import dirname, join, exists, expanduser
from os import makedirs
import json
from pprint import pprint



# -*- coding: utf-8 -*-
#
# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime

from dateutil.relativedelta import relativedelta

from tnalagmes.lang.parse_common import is_numeric, look_for_fractions
from tnalagmes.lang.format_en import NUM_STRING_EN, LONG_SCALE_EN, \
    SHORT_SCALE_EN

SHORT_ORDINAL_STRING_EN = {
    1: 'first',
    2: 'second',
    3: 'third',
    4: 'fourth',
    5: 'fifth',
    6: 'sixth',
    7: 'seventh',
    8: 'eighth',
    9: 'ninth',
    10: 'tenth',
    11: 'eleventh',
    12: 'twelfth',
    13: 'thirteenth',
    14: 'fourteenth',
    15: 'fifteenth',
    16: 'sixteenth',
    17: 'seventeenth',
    18: 'eighteenth',
    19: 'nineteenth',
    20: 'twentieth',
    30: 'thirtieth',
    40: "fortieth",
    50: "fiftieth",
    60: "sixtieth",
    70: "seventieth",
    80: "eightieth",
    90: "ninetieth",
    10e-3: "hundredth",
    1e-3: "thousandth",
    1e-6: "millionth",
    1e-9: "billionth",
    1e-12: "trillionth",
    1e-15: "quadrillionth",
    1e-18: "quintillionth",
    1e-21: "sextillionth",
    1e-24: "septillionth",
    1e-27: "octillionth",
    1e-30: "nonillionth",
    1e-33: "decillionth"
    # TODO > 1e-33
}

LONG_ORDINAL_STRING_EN = {
    1: 'first',
    2: 'second',
    3: 'third',
    4: 'fourth',
    5: 'fifth',
    6: 'sixth',
    7: 'seventh',
    8: 'eighth',
    9: 'ninth',
    10: 'tenth',
    11: 'eleventh',
    12: 'twelfth',
    13: 'thirteenth',
    14: 'fourteenth',
    15: 'fifteenth',
    16: 'sixteenth',
    17: 'seventeenth',
    18: 'eighteenth',
    19: 'nineteenth',
    20: 'twentieth',
    30: 'thirtieth',
    40: "fortieth",
    50: "fiftieth",
    60: "sixtieth",
    70: "seventieth",
    80: "eightieth",
    90: "ninetieth",
    10e-3: "hundredth",
    1e-3: "thousandth",
    1e-6: "millionth",
    1e-12: "billionth",
    1e-18: "trillionth",
    1e-24: "quadrillionth",
    1e-30: "quintillionth",
    1e-36: "sextillionth",
    1e-42: "septillionth",
    1e-48: "octillionth",
    1e-54: "nonillionth",
    1e-60: "decillionth"
    # TODO > 1e-60
}


def extract_number_en(text, short_scale=True, ordinals=False):
    """
    This function extracts a number from a text string,
    handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """

    def _normalize(text):
        text = text.lower()
        erases = ["the", "of", "a", "an", "to", "positive", "plus"]
        replaces = {
            "exponentiated": "power",
            "raised": "power",
            "elevated": "power",
            "by": "times"  # scientific notation
        }
        check_duplicates = ["power"]
        # cardinals
        if short_scale:
            cards = [SHORT_ORDINAL_STRING_EN[c]
                     for c in SHORT_ORDINAL_STRING_EN.keys()]
        else:
            cards = [LONG_ORDINAL_STRING_EN[c]
                     for c in LONG_ORDINAL_STRING_EN.keys()]
        words = text.split(" ")
        for idx, word in enumerate(words):
            prev_word = words[idx - 1] if idx > 0 else ""
            if word == "power" and prev_word in cards:
                i = cards.index(prev_word) + 1
                # TODO > 20
                if i <= 20:
                    words[idx - 1] = NUM_STRING_EN[i]
            elif prev_word == "power" and word in cards:
                i = cards.index(word) + 1
                # TODO > 20
                if i <= 20:
                    words[idx] = word = NUM_STRING_EN[i]
            if word in erases:
                words[idx] = ""
            elif word in replaces.keys():
                words[idx] = replaces[word]
                if replaces[word] in check_duplicates and \
                        replaces[word] in " ".join(words[:idx]):
                    words[idx] = ""
            if word in check_duplicates and word in " ".join(words[:idx]):
                words[idx] = ""

        return " ".join(words).rstrip().lstrip()

    text = _normalize(text)
    string_num_en = {
                     "half": 0.5,
                     "halves": 0.5,
                     "hundreds": 100,
                     "thousands": 1000,
                     'millions': 1000000}

    for num in NUM_STRING_EN:
        num_string = NUM_STRING_EN[num]
        string_num_en[num_string] = num

    # first, second...
    if ordinals:
        if short_scale:
            for num in SHORT_ORDINAL_STRING_EN:
                num_string = SHORT_ORDINAL_STRING_EN[num]
                string_num_en[num_string] = num
        else:
            for num in LONG_ORDINAL_STRING_EN:
                num_string = LONG_ORDINAL_STRING_EN[num]
                string_num_en[num_string] = num

    # negate next number (-2 = 0 - 2)
    negatives = ["negative", "minus"]

    # sum the next number (twenty two = 20 + 2)
    sums = ['twenty', 'thirty', 'forty', 'fifty', 'sixty', 'seventy', 'eighty',
            'ninety']

    # multiply the previous number (one hundred = 1 * 100)
    multiplies = ["hundred", "thousand", "hundreds", "thousands", "million",
                  "millions"]

    # split sentence parse separately and sum ( 2 and a half = 2 + 0.5 )
    fraction_marker = [" and "]

    # decimal marker ( 1 point 5 = 1 + 0.5)
    decimal_marker = [" point ", " dot "]

    if short_scale:
        for num in SHORT_SCALE_EN:
            num_string = SHORT_SCALE_EN[num]
            string_num_en[num_string] = num
            string_num_en[num_string + "s"] = num
            multiplies.append(num_string)
            multiplies.append(num_string + "s")
    else:
        for num in LONG_SCALE_EN:
            num_string = LONG_SCALE_EN[num]
            string_num_en[num_string] = num
            string_num_en[num_string + "s"] = num
            multiplies.append(num_string)
            multiplies.append(num_string + "s")

    # 2 and 3/4
    for c in fraction_marker:
        components = text.split(c)

        if len(components) == 2:
            # ensure first is not a fraction and second is a fraction
            num1 = extract_number_en(components[0])
            num2 = extract_number_en(components[1])
            if num1 is not False and num2 is not False \
                    and num1 >= 1 and 0 < num2 < 1:
                return num1 + num2

    # 2 point 5
    for c in decimal_marker:
        components = text.split(c)
        if len(components) == 2:
            number = extract_number_en(components[0])
            decimal = extract_number_en(components[1])
            if number is not False and decimal is not False:
                # TODO handle number dot number number number
                if "." not in str(decimal):
                    return number + float("0." + str(decimal))

    aWords = text.split()
    aWords = [word for word in aWords if word not in ["the", "a", "an"]]
    val = False
    prev_val = None
    to_sum = []
    for idx, word in enumerate(aWords):

        if not word:
            continue
        prev_word = aWords[idx - 1] if idx > 0 else ""
        next_word = aWords[idx + 1] if idx + 1 < len(aWords) else ""

        # is this word already a number ?
        if is_numeric(word):
            # if word.isdigit():            # doesn't work with decimals
            val = float(word)

        # is this word the name of a number ?
        if word in string_num_en:
            val = string_num_en[word]

        # is the prev word a number and should we sum it?
        # twenty two, fifty six
        if prev_word in sums and word in string_num_en:
            if val and val < 10:
                val = prev_val + val

        # is the prev word a number and should we multiply it?
        # twenty hundred, six hundred
        if word in multiplies:
            if not prev_val:
                prev_val = 1
            val = prev_val * val

        # is this a spoken fraction?
        # half cup
        if val is False:
            val = is_fractional_en(word, short_scale=short_scale)

        # 2 fifths
        if not ordinals:
            next_value = is_fractional_en(next_word, short_scale=short_scale)
            if next_value:
                if not val:
                    val = 1
                val = val * next_value

        # is this a negative number?
        if val and prev_word and prev_word in negatives:
            val = 0 - val

        # let's make sure it isn't a fraction
        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])

        else:
            prev_val = val
            # handle long numbers
            # six hundred sixty six
            # two million five hundred thousand
            if word in multiplies and next_word not in multiplies:
                to_sum.append(val)
                val = 0
                prev_val = 0
            # scientific notation
            elif prev_word == "times" and \
                    word in ["ten", "10"] and next_word == "power":

                power = int(extract_number_en(" ".join(aWords[idx:])))
                val = extract_number_en(" ".join(aWords[:idx]))
                if val:
                    return float(str(val) + "e" + str(power))
            elif prev_word == "times" and \
                    word in ["ten", "10"] and \
                    extract_number_en(" ".join(aWords[idx:])) \
                    and text.endswith("power"):

                power = int(extract_number_en(" ".join(aWords[idx:])))
                val = extract_number_en(" ".join(aWords[:idx]))
                if val:
                    return float(str(val) + "e" + str(power))

    if val is not None:
        for v in to_sum:
            val = val + v
    return val


def extract_datetime_en(string, currentDate):
    """ Convert a human date reference into an exact datetime

    Convert things like
        "today"
        "tomorrow afternoon"
        "next Tuesday at 4pm"
        "August 3rd"
    into a datetime.  If a reference date is not provided, the current
    local time is used.  Also consumes the words used to define the date
    returning the remaining string.  For example, the string
       "what is Tuesday's weather forecast"
    returns the date for the forthcoming Tuesday relative to the reference
    date and the remainder string
       "what is weather forecast".

    Args:
        string (str): string containing date words
        currentDate (datetime): A reference date/time for "tommorrow", etc

    Returns:
        [datetime, str]: An array containing the datetime and the remaining
                         text not consumed in the parsing.
    """

    def clean_string(s):
        # clean unneeded punctuation and capitalization among other things.
        s = s.lower().replace('?', '').replace('.', '').replace(',', '') \
            .replace(' the ', ' ').replace(' a ', ' ').replace(' an ', ' ') \
            .replace("o' clock", "o'clock").replace("o clock", "o'clock") \
            .replace("o ' clock", "o'clock").replace("o 'clock", "o'clock")

        wordList = s.split()
        for idx, word in enumerate(wordList):
            word = word.replace("'s", "")

            ordinals = ["rd", "st", "nd", "th"]
            if word[0].isdigit():
                for ordinal in ordinals:
                    # "second" is the only case we should not do this
                    if ordinal in word and "second" not in word:
                        word = word.replace(ordinal, "")
            wordList[idx] = word

        return wordList

    def date_found():
        return found or \
               (
                       datestr != "" or timeStr != "" or
                       yearOffset != 0 or monthOffset != 0 or
                       dayOffset is True or hrOffset != 0 or
                       hrAbs != 0 or minOffset != 0 or
                       minAbs != 0 or secOffset != 0
               )

    if string == "" or not currentDate:
        return None

    found = False
    daySpecified = False
    dayOffset = False
    monthOffset = 0
    yearOffset = 0
    dateNow = currentDate
    today = dateNow.strftime("%w")
    currentYear = dateNow.strftime("%Y")
    fromFlag = False
    datestr = ""
    hasYear = False
    timeQualifier = ""

    timeQualifiersAM = ['morning']
    timeQualifiersPM = ['afternoon', 'evening', 'tonight', 'night']
    timeQualifiersList = set(timeQualifiersAM + timeQualifiersPM)
    markers = ['at', 'in', 'on', 'by', 'this', 'around', 'for', 'of']
    days = ['monday', 'tuesday', 'wednesday',
            'thursday', 'friday', 'saturday', 'sunday']
    months = ['january', 'february', 'march', 'april', 'may', 'june',
              'july', 'august', 'september', 'october', 'november',
              'december']
    monthsShort = ['jan', 'feb', 'mar', 'apr', 'may', 'june', 'july', 'aug',
                   'sept', 'oct', 'nov', 'dec']

    words = clean_string(string)

    for idx, word in enumerate(words):
        if word == "":
            continue
        wordPrevPrev = words[idx - 2] if idx > 1 else ""
        wordPrev = words[idx - 1] if idx > 0 else ""
        wordNext = words[idx + 1] if idx + 1 < len(words) else ""
        wordNextNext = words[idx + 2] if idx + 2 < len(words) else ""

        # this isn't in clean string because I don't want to save back to words
        word = word.rstrip('s')
        start = idx
        used = 0
        # save timequalifier for later
        if word in timeQualifiersList:
            timeQualifier = word
            # parse today, tomorrow, day after tomorrow
        elif word == "today" and not fromFlag:
            dayOffset = 0
            used += 1
        elif word == "tomorrow" and not fromFlag:
            dayOffset = 1
            used += 1
        elif (word == "day" and
              wordNext == "after" and
              wordNextNext == "tomorrow" and
              not fromFlag and
              not wordPrev[0].isdigit()):
            dayOffset = 2
            used = 3
            if wordPrev == "the":
                start -= 1
                used += 1
                # parse 5 days, 10 weeks, last week, next week
        elif word == "day":
            if wordPrev[0].isdigit():
                dayOffset += int(wordPrev)
                start -= 1
                used = 2
        elif word == "week" and not fromFlag:
            if wordPrev[0].isdigit():
                dayOffset += int(wordPrev) * 7
                start -= 1
                used = 2
            elif wordPrev == "next":
                dayOffset = 7
                start -= 1
                used = 2
            elif wordPrev == "last":
                dayOffset = -7
                start -= 1
                used = 2
                # parse 10 months, next month, last month
        elif word == "month" and not fromFlag:
            if wordPrev[0].isdigit():
                monthOffset = int(wordPrev)
                start -= 1
                used = 2
            elif wordPrev == "next":
                monthOffset = 1
                start -= 1
                used = 2
            elif wordPrev == "last":
                monthOffset = -1
                start -= 1
                used = 2
                # parse 5 years, next year, last year
        elif word == "year" and not fromFlag:
            if wordPrev[0].isdigit():
                yearOffset = int(wordPrev)
                start -= 1
                used = 2
            elif wordPrev == "next":
                yearOffset = 1
                start -= 1
                used = 2
            elif wordPrev == "last":
                yearOffset = -1
                start -= 1
                used = 2
                # parse Monday, Tuesday, etc., and next Monday,
                # last Tuesday, etc.
        elif word in days and not fromFlag:
            d = days.index(word)
            dayOffset = (d + 1) - int(today)
            used = 1
            if dayOffset < 0:
                dayOffset += 7
            if wordPrev == "next":
                dayOffset += 7
                used += 1
                start -= 1
            elif wordPrev == "last":
                dayOffset -= 7
                used += 1
                start -= 1
                # parse 15 of July, June 20th, Feb 18, 19 of February
        elif word in months or word in monthsShort and not fromFlag:
            try:
                m = months.index(word)
            except ValueError:
                m = monthsShort.index(word)
            used += 1
            datestr = months[m]
            if wordPrev and (wordPrev[0].isdigit() or
                             (wordPrev == "of" and wordPrevPrev[0].isdigit())):
                if wordPrev == "of" and wordPrevPrev[0].isdigit():
                    datestr += " " + words[idx - 2]
                    used += 1
                    start -= 1
                else:
                    datestr += " " + wordPrev
                start -= 1
                used += 1
                if wordNext and wordNext[0].isdigit():
                    datestr += " " + wordNext
                    used += 1
                    hasYear = True
                else:
                    hasYear = False

            elif wordNext and wordNext[0].isdigit():
                datestr += " " + wordNext
                used += 1
                if wordNextNext and wordNextNext[0].isdigit():
                    datestr += " " + wordNextNext
                    used += 1
                    hasYear = True
                else:
                    hasYear = False
        # parse 5 days from tomorrow, 10 weeks from next thursday,
        # 2 months from July
        validFollowups = days + months + monthsShort
        validFollowups.append("today")
        validFollowups.append("tomorrow")
        validFollowups.append("next")
        validFollowups.append("last")
        validFollowups.append("now")
        if (word == "from" or word == "after") and wordNext in validFollowups:
            used = 2
            fromFlag = True
            if wordNext == "tomorrow":
                dayOffset += 1
            elif wordNext in days:
                d = days.index(wordNext)
                tmpOffset = (d + 1) - int(today)
                used = 2
                if tmpOffset < 0:
                    tmpOffset += 7
                dayOffset += tmpOffset
            elif wordNextNext and wordNextNext in days:
                d = days.index(wordNextNext)
                tmpOffset = (d + 1) - int(today)
                used = 3
                if wordNext == "next":
                    tmpOffset += 7
                    used += 1
                    start -= 1
                elif wordNext == "last":
                    tmpOffset -= 7
                    used += 1
                    start -= 1
                dayOffset += tmpOffset
        if used > 0:
            if start - 1 > 0 and words[start - 1] == "this":
                start -= 1
                used += 1

            for i in range(0, used):
                words[i + start] = ""

            if start - 1 >= 0 and words[start - 1] in markers:
                words[start - 1] = ""
            found = True
            daySpecified = True

    # parse time
    timeStr = ""
    hrOffset = 0
    minOffset = 0
    secOffset = 0
    hrAbs = 0
    minAbs = 0
    military = False

    for idx, word in enumerate(words):
        if word == "":
            continue

        wordPrevPrev = words[idx - 2] if idx > 1 else ""
        wordPrev = words[idx - 1] if idx > 0 else ""
        wordNext = words[idx + 1] if idx + 1 < len(words) else ""
        wordNextNext = words[idx + 2] if idx + 2 < len(words) else ""
        # parse noon, midnight, morning, afternoon, evening
        used = 0
        if word == "noon":
            hrAbs = 12
            used += 1
        elif word == "midnight":
            hrAbs = 0
            used += 1
        elif word == "morning":
            if hrAbs == 0:
                hrAbs = 8
            used += 1
        elif word == "afternoon":
            if hrAbs == 0:
                hrAbs = 15
            used += 1
        elif word == "evening":
            if hrAbs == 0:
                hrAbs = 19
            used += 1
            # parse half an hour, quarter hour
        elif word == "hour" and \
                (wordPrev in markers or wordPrevPrev in markers):
            if wordPrev == "half":
                minOffset = 30
            elif wordPrev == "quarter":
                minOffset = 15
            elif wordPrevPrev == "quarter":
                minOffset = 15
                if idx > 2 and words[idx - 3] in markers:
                    words[idx - 3] = ""
                    if words[idx - 3] == "this":
                        daySpecified = True
                words[idx - 2] = ""
            else:
                hrOffset = 1
            if wordPrevPrev in markers:
                words[idx - 2] = ""
                if wordPrevPrev == "this":
                    daySpecified = True
            words[idx - 1] = ""
            used += 1
            hrAbs = -1
            minAbs = -1
            # parse 5:00 am, 12:00 p.m., etc
        elif word[0].isdigit():
            isTime = True
            strHH = ""
            strMM = ""
            remainder = ""
            if ':' in word:
                # parse colons
                # "3:00 in the morning"
                stage = 0
                length = len(word)
                for i in range(length):
                    if stage == 0:
                        if word[i].isdigit():
                            strHH += word[i]
                        elif word[i] == ":":
                            stage = 1
                        else:
                            stage = 2
                            i -= 1
                    elif stage == 1:
                        if word[i].isdigit():
                            strMM += word[i]
                        else:
                            stage = 2
                            i -= 1
                    elif stage == 2:
                        remainder = word[i:].replace(".", "")
                        break
                if remainder == "":
                    nextWord = wordNext.replace(".", "")
                    if nextWord == "am" or nextWord == "pm":
                        remainder = nextWord
                        used += 1
                    elif nextWord == "tonight":
                        remainder = "pm"
                        used += 1
                    elif wordNext == "in" and wordNextNext == "the" and \
                            words[idx + 3] == "morning":
                        remainder = "am"
                        used += 3
                    elif wordNext == "in" and wordNextNext == "the" and \
                            words[idx + 3] == "afternoon":
                        remainder = "pm"
                        used += 3
                    elif wordNext == "in" and wordNextNext == "the" and \
                            words[idx + 3] == "evening":
                        remainder = "pm"
                        used += 3
                    elif wordNext == "in" and wordNextNext == "morning":
                        remainder = "am"
                        used += 2
                    elif wordNext == "in" and wordNextNext == "afternoon":
                        remainder = "pm"
                        used += 2
                    elif wordNext == "in" and wordNextNext == "evening":
                        remainder = "pm"
                        used += 2
                    elif wordNext == "this" and wordNextNext == "morning":
                        remainder = "am"
                        used = 2
                        daySpecified = True
                    elif wordNext == "this" and wordNextNext == "afternoon":
                        remainder = "pm"
                        used = 2
                        daySpecified = True
                    elif wordNext == "this" and wordNextNext == "evening":
                        remainder = "pm"
                        used = 2
                        daySpecified = True
                    elif wordNext == "at" and wordNextNext == "night":
                        if strHH and int(strHH) > 5:
                            remainder = "pm"
                        else:
                            remainder = "am"
                        used += 2
                    else:
                        if timeQualifier != "":
                            military = True
                            if strHH and int(strHH) <= 12 and \
                                    (timeQualifier in timeQualifiersPM):
                                strHH += str(int(strHH) + 12)
            else:
                # try to parse numbers without colons
                # 5 hours, 10 minutes etc.
                length = len(word)
                strNum = ""
                remainder = ""
                for i in range(length):
                    if word[i].isdigit():
                        strNum += word[i]
                    else:
                        remainder += word[i]

                if remainder == "":
                    remainder = wordNext.replace(".", "").lstrip().rstrip()
                if (
                        remainder == "pm" or
                        wordNext == "pm" or
                        remainder == "p.m." or
                        wordNext == "p.m."):
                    strHH = strNum
                    remainder = "pm"
                    used = 1
                elif (
                        remainder == "am" or
                        wordNext == "am" or
                        remainder == "a.m." or
                        wordNext == "a.m."):
                    strHH = strNum
                    remainder = "am"
                    used = 1
                else:
                    if (
                            int(strNum) > 100 and
                            (
                                    wordPrev == "o" or
                                    wordPrev == "oh"
                            )):
                        # 0800 hours (pronounced oh-eight-hundred)
                        strHH = str(int(strNum) // 100)
                        strMM = str(int(strNum) % 100)
                        military = True
                        if wordNext == "hours":
                            used += 1
                    elif (
                            (wordNext == "hours" or wordNext == "hour" or
                             remainder == "hours" or remainder == "hour") and
                            word[0] != '0' and
                            (
                                    int(strNum) < 100 or
                                    int(strNum) > 2400
                            )):
                        # ignores military time
                        # "in 3 hours"
                        hrOffset = int(strNum)
                        used = 2
                        isTime = False
                        hrAbs = -1
                        minAbs = -1

                    elif wordNext == "minutes" or wordNext == "minute" or \
                            remainder == "minutes" or remainder == "minute":
                        # "in 10 minutes"
                        minOffset = int(strNum)
                        used = 2
                        isTime = False
                        hrAbs = -1
                        minAbs = -1
                    elif wordNext == "seconds" or wordNext == "second" \
                            or remainder == "seconds" or remainder == "second":
                        # in 5 seconds
                        secOffset = int(strNum)
                        used = 2
                        isTime = False
                        hrAbs = -1
                        minAbs = -1
                    elif int(strNum) > 100:
                        # military time, eg. "3300 hours"
                        strHH = str(int(strNum) // 100)
                        strMM = str(int(strNum) % 100)
                        military = True
                        if wordNext == "hours" or wordNext == "hour" or \
                                remainder == "hours" or remainder == "hour":
                            used += 1
                    elif wordNext and wordNext[0].isdigit():
                        # military time, e.g. "04 38 hours"
                        strHH = strNum
                        strMM = wordNext
                        military = True
                        used += 1
                        if (wordNextNext == "hours" or
                                wordNextNext == "hour" or
                                remainder == "hours" or remainder == "hour"):
                            used += 1
                    elif (
                            wordNext == "" or wordNext == "o'clock" or
                            (
                                    wordNext == "in" and
                                    (
                                            wordNextNext == "the" or
                                            wordNextNext == timeQualifier
                                    )
                            )):
                        strHH = strNum
                        strMM = "00"
                        if wordNext == "o'clock":
                            used += 1
                        if wordNext == "in" or wordNextNext == "in":
                            used += (1 if wordNext == "in" else 2)
                            wordNextNextNext = words[idx + 3] \
                                if idx + 3 < len(words) else ""

                            if (wordNextNext and
                                    (wordNextNext in timeQualifier or
                                     wordNextNextNext in timeQualifier)):
                                if (wordNextNext in timeQualifiersPM or
                                        wordNextNextNext in timeQualifiersPM):
                                    remainder = "pm"
                                    used += 1
                                if (wordNextNext in timeQualifiersAM or
                                        wordNextNextNext in timeQualifiersAM):
                                    remainder = "am"
                                    used += 1
                        if timeQualifier != "":
                            used += 1  # TODO: Unsure if this is 100% accurate
                            military = True
                    else:
                        isTime = False

            HH = int(strHH) if strHH else 0
            MM = int(strMM) if strMM else 0
            HH = HH + 12 if remainder == "pm" and HH < 12 else HH
            HH = HH - 12 if remainder == "am" and HH >= 12 else HH

            if not military and not (remainder in ['am', 'pm']):
                if not daySpecified or dayOffset < 1:
                    # ambiguous time, detect whether they mean this evening or
                    # the next morning based on whether it has already passed
                    if HH > dateNow.hour:
                        # has passed, assume the next morning
                        dayOffset += 1
                    elif HH <= 12:
                        # forthcoming, assume this afternoon/evening
                        HH += 12

            if timeQualifier in timeQualifiersPM and HH < 12:
                HH += 12

            if HH > 24 or MM > 59:
                isTime = False
                used = 0
            if isTime:
                hrAbs = HH
                minAbs = MM
                used += 1
        if used > 0:
            # removed parsed words from the sentence
            for i in range(used):
                if idx + i >= len(words):
                    break
                words[idx + i] = ""

            if wordPrev == "o" or wordPrev == "oh":
                words[words.index(wordPrev)] = ""

            if wordPrev == "early":
                hrOffset = -1
                words[idx - 1] = ""
                idx -= 1
            elif wordPrev == "late":
                hrOffset = 1
                words[idx - 1] = ""
                idx -= 1
            if idx > 0 and wordPrev in markers:
                words[idx - 1] = ""
                if wordPrev == "this":
                    daySpecified = True
            if idx > 1 and wordPrevPrev in markers:
                words[idx - 2] = ""
                if wordPrevPrev == "this":
                    daySpecified = True

            idx += used - 1
            found = True

    # check that we found a date
    if not date_found:
        return None

    if dayOffset is False:
        dayOffset = 0

    # perform date manipulation

    extractedDate = dateNow.replace(microsecond=0)

    if datestr != "":
        # date included an explicit date, e.g. "june 5" or "june 2, 2017"
        try:
            temp = datetime.strptime(datestr, "%B %d")
        except ValueError:
            # Try again, allowing the year
            temp = datetime.strptime(datestr, "%B %d %Y")
        extractedDate = extractedDate.replace(hour=0, minute=0, second=0)
        if not hasYear:
            temp = temp.replace(year=extractedDate.year,
                                tzinfo=extractedDate.tzinfo)
            if extractedDate < temp:
                extractedDate = extractedDate.replace(
                    year=int(currentYear),
                    month=int(temp.strftime("%m")),
                    day=int(temp.strftime("%d")),
                    tzinfo=extractedDate.tzinfo)
            else:
                extractedDate = extractedDate.replace(
                    year=int(currentYear) + 1,
                    month=int(temp.strftime("%m")),
                    day=int(temp.strftime("%d")),
                    tzinfo=extractedDate.tzinfo)
        else:
            extractedDate = extractedDate.replace(
                year=int(temp.strftime("%Y")),
                month=int(temp.strftime("%m")),
                day=int(temp.strftime("%d")),
                tzinfo=extractedDate.tzinfo)
    else:
        # ignore the current HH:MM:SS if relative using days or greater
        if not timeStr and hrOffset == 0 and minOffset == 0 and secOffset == 0:
            extractedDate = extractedDate.replace(hour=0, minute=0, second=0)

    if timeStr != "":
        temp = datetime(timeStr)
        extractedDate = extractedDate.replace(hour=temp.strftime("%H"),
                                              minute=temp.strftime("%M"),
                                              second=temp.strftime("%S"),
                                              tzinfo=extractedDate.tzinfo)

    if yearOffset != 0:
        extractedDate = extractedDate + relativedelta(years=yearOffset)
    if monthOffset != 0:
        extractedDate = extractedDate + relativedelta(months=monthOffset)
    if dayOffset != 0:
        extractedDate = extractedDate + relativedelta(days=dayOffset)
    if hrAbs != -1 and minAbs != -1:

        extractedDate = extractedDate + relativedelta(hours=hrAbs,
                                                      minutes=minAbs)
        if (hrAbs != 0 or minAbs != 0) and datestr == "":
            if not daySpecified and dateNow > extractedDate:
                extractedDate = extractedDate + relativedelta(days=1)
    if hrOffset != 0:
        extractedDate = extractedDate + relativedelta(hours=hrOffset)
    if minOffset != 0:
        extractedDate = extractedDate + relativedelta(minutes=minOffset)
    if secOffset != 0:
        extractedDate = extractedDate + relativedelta(seconds=secOffset)
    for idx, word in enumerate(words):
        if words[idx] == "and" and words[idx - 1] == "" and words[
            idx + 1] == "":
            words[idx] = ""

    resultStr = " ".join(words)
    resultStr = ' '.join(resultStr.split())
    return [extractedDate, resultStr]


def is_fractional_en(input_str, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    if input_str.endswith('s', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "fifths"

    fracts = {"whole": 1, "half": 2, "halve": 2, "quarter": 4}
    if short_scale:
        for num in SHORT_ORDINAL_STRING_EN:
            if num > 2:
                fracts[SHORT_ORDINAL_STRING_EN[num]] = num
    else:
        for num in LONG_ORDINAL_STRING_EN:
            if num > 2:
                fracts[LONG_ORDINAL_STRING_EN[num]] = num

    if input_str.lower() in fracts:
        return 1.0 / fracts[input_str.lower()]
    return False


def normalize_en(text, remove_articles):
    """ English string normalization """

    words = text.split()  # this also removed extra spaces
    normalized = ""
    for word in words:
        if remove_articles and word in ["the", "a", "an"]:
            continue

        # Expand common contractions, e.g. "isn't" -> "is not"
        contraction = ["ain't", "aren't", "can't", "could've", "couldn't",
                       "didn't", "doesn't", "don't", "gonna", "gotta",
                       "hadn't", "hasn't", "haven't", "he'd", "he'll", "he's",
                       "how'd", "how'll", "how's", "I'd", "I'll", "I'm",
                       "I've", "isn't", "it'd", "it'll", "it's", "mightn't",
                       "might've", "mustn't", "must've", "needn't",
                       "oughtn't",
                       "shan't", "she'd", "she'll", "she's", "shouldn't",
                       "should've", "somebody's", "someone'd", "someone'll",
                       "someone's", "that'll", "that's", "that'd", "there'd",
                       "there're", "there's", "they'd", "they'll", "they're",
                       "they've", "wasn't", "we'd", "we'll", "we're", "we've",
                       "weren't", "what'd", "what'll", "what're", "what's",
                       "whats",  # technically incorrect but some STT outputs
                       "what've", "when's", "when'd", "where'd", "where's",
                       "where've", "who'd", "who'd've", "who'll", "who're",
                       "who's", "who've", "why'd", "why're", "why's", "won't",
                       "won't've", "would've", "wouldn't", "wouldn't've",
                       "y'all", "ya'll", "you'd", "you'd've", "you'll",
                       "y'aint", "y'ain't", "you're", "you've"]
        if word in contraction:
            expansion = ["is not", "are not", "can not", "could have",
                         "could not", "did not", "does not", "do not",
                         "going to", "got to", "had not", "has not",
                         "have not", "he would", "he will", "he is",
                         "how did",
                         "how will", "how is", "I would", "I will", "I am",
                         "I have", "is not", "it would", "it will", "it is",
                         "might not", "might have", "must not", "must have",
                         "need not", "ought not", "shall not", "she would",
                         "she will", "she is", "should not", "should have",
                         "somebody is", "someone would", "someone will",
                         "someone is", "that will", "that is", "that would",
                         "there would", "there are", "there is", "they would",
                         "they will", "they are", "they have", "was not",
                         "we would", "we will", "we are", "we have",
                         "were not", "what did", "what will", "what are",
                         "what is",
                         "what is", "what have", "when is", "when did",
                         "where did", "where is", "where have", "who would",
                         "who would have", "who will", "who are", "who is",
                         "who have", "why did", "why are", "why is",
                         "will not", "will not have", "would have",
                         "would not", "would not have", "you all", "you all",
                         "you would", "you would have", "you will",
                         "you are not", "you are not", "you are", "you have"]
            word = expansion[contraction.index(word)]

        # Convert numbers into digits, e.g. "two" -> "2"
        textNumbers = ["zero", "one", "two", "three", "four", "five", "six",
                       "seven", "eight", "nine", "ten", "eleven", "twelve",
                       "thirteen", "fourteen", "fifteen", "sixteen",
                       "seventeen", "eighteen", "nineteen", "twenty"]

        if word in textNumbers:
            word = str(textNumbers.index(word))

        normalized += " " + word

    return normalized[1:]  # strip the initial space




# -*- coding: utf-8 -*-
#
# Copyright 2018 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from datetime import datetime
from dateutil.tz import gettz, tzlocal


def default_timezone():
    """ Get the default timezone

    Based on user location settings location.timezone.code or
    the default system value if no setting exists.

    Returns:
        (datetime.tzinfo): Definition of the default timezone
    """
    # Just go with system default timezone
    return tzlocal()


def now_utc():
    """ Retrieve the current time in UTC

    Returns:
        (datetime): The current time in Universal Time, aka GMT
    """
    return datetime.utcnow()


def now_local(tz=None):
    """ Retrieve the current time

    Args:
        tz (datetime.tzinfo, optional): Timezone, default to user's settings

    Returns:
        (datetime): The current time
    """
    if not tz:
        tz = default_timezone()
    return datetime.now(tz)


def to_utc(dt):
    """ Convert a datetime with timezone info to a UTC datetime

    Args:
        dt (datetime): A datetime (presumably in some local zone)
    Returns:
        (datetime): time converted to UTC
    """
    tzUTC = gettz("UTC")
    if dt.tzinfo:
        return dt.astimezone(tzUTC)
    else:
        return dt.replace(tzinfo=gettz("UTC")).astimezone(tzUTC)


def to_local(dt):
    """ Convert a datetime to the user's local timezone

    Args:
        dt (datetime): A datetime (if no timezone, defaults to UTC)
    Returns:
        (datetime): time converted to the local timezone
    """
    tz = default_timezone()
    if dt.tzinfo:
        return dt.astimezone(tz)
    else:
        return dt.replace(tzinfo=gettz("UTC")).astimezone(tz)


def get_timedelta(dt, anchor=None):
    """ Get a datetime object or a int() Epoch timestamp and return a timedelta"""
    anchor = anchor or now_local()
    if type(dt) is int:
        dt = datetime.fromtimestamp(dt, tz=default_timezone())
    if type(anchor) is int:
        anchor = datetime.fromtimestamp(anchor)
    anchor = to_local(anchor)
    dt = to_local(dt)
    return anchor - dt



# -*- coding: utf-8 -*-
#
# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from tnalagmes.lang.format_common import convert_to_mixed_fraction

NUM_STRING_IT = {
    0: 'zero',
    1: 'uno',
    2: 'due',
    3: 'tre',
    4: 'quattro',
    5: 'cinque',
    6: 'sei',
    7: 'sette',
    8: 'otto',
    9: 'nove',
    10: 'dieci',
    11: 'undici',
    12: 'dodici',
    13: 'tredici',
    14: 'quattordici',
    15: 'quindici',
    16: 'sedici',
    17: 'diciassette',
    18: 'diciotto',
    19: 'diciannove',
    20: 'venti',
    30: 'trenta',
    40: 'quaranta',
    50: 'cinquanta',
    60: 'sessanta',
    70: 'settanta',
    80: 'ottanta',
    90: 'novanta'
}

FRACTION_STRING_IT = {
    2: 'mezz',
    3: 'terz',
    4: 'quart',
    5: 'quint',
    6: 'sest',
    7: 'settim',
    8: 'ottav',
    9: 'non',
    10: 'decim',
    11: 'undicesim',
    12: 'dodicesim',
    13: 'tredicesim',
    14: 'quattordicesim',
    15: 'quindicesim',
    16: 'sedicesim',
    17: 'diciassettesim',
    18: 'diciottesim',
    19: 'diciannovesim',
    20: 'ventesim'
}


def nice_number_it(number, speech, denominators):
    """ Italian helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 e un mezz" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    # denominatore
    den_str = FRACTION_STRING_IT[den]
    # frazione
    if whole == 0:
        if num == 1:
            # un decimo
            return_string = 'un {}'.format(den_str)
        else:
            # tre mezzi
            return_string = '{} {}'.format(num, den_str)
    # interi  >10
    elif num == 1:
        # trenta e un
        return_string = '{} e un {}'.format(whole, den_str)
    # interi >10 con frazioni
    else:
        # venti e 3 decimi
        return_string = '{} e {} {}'.format(whole, num, den_str)

    # gestisce il plurale del denominatore
    if num > 1:
        return_string += 'i'
    else:
        return_string += 'o'

    return return_string


def pronounce_number_it(num, places=2):
    """
    Convert a number to it's spoken equivalent
    adapted to italian fron en version

    For example, '5.2' would return 'cinque virgola due'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
    Returns:
        (str): The pronounced number
    """
    if abs(num) >= 100:
        # TODO: Support for numbers over 100
        return str(num)

    result = ""
    if num < 0:
        result = "meno "
    num = abs(num)

    if num > 20:
        tens = int(num - int(num) % 10)
        ones = int(num - tens)
        result += NUM_STRING_IT[tens]
        if ones > 0:
            if ones == 1 or ones == 8:
                result = result[:-1]  # ventuno  ventotto
            result += NUM_STRING_IT[ones]

    else:
        result += NUM_STRING_IT[int(num)]

    # Deal with fractional part
    if not num == int(num) and places > 0:
        result += " virgola"
        place = 10
        while int(num * place) % 10 > 0 and places > 0:
            result += " " + NUM_STRING_IT[int(num * place) % 10]
            place *= 10
            places -= 1
    return result


def nice_time_it(dt, speech=True, use_24hour=False, use_ampm=False):
    """
    Format a time to a comfortable human format
    adapted to italian fron en version

    For example, generate 'cinque e trenta' for speech or '5:30' for
    text display.

    Args:
        dt (datetime): date to format (assumes already in local timezone)
        speech (bool): format for speech (default/True) or display (False)=Fal
        use_24hour (bool): output in 24-hour/military or 12-hour format
        use_ampm (bool): include the am/pm for 12-hour format
    Returns:
        (str): The formatted time string
    """
    if use_24hour:
        # e.g. "03:01" or "14:22"
        string = dt.strftime("%H:%M")
    else:
        if use_ampm:
            # e.g. "3:01 AM" or "2:22 PM"
            string = dt.strftime("%I:%M %p")
        else:
            # e.g. "3:01" or "2:22"
            string = dt.strftime("%I:%M")
        if string[0] == '0':
            string = string[1:]  # strip leading zeros

    if not speech:
        return string

    # Generate a speakable version of the time
    if use_24hour:
        speak = ""
        # Either "zero 8 zerozero" o "13 zerozero"
        if string[0:2] == '00':
            speak += "zerozero"
        elif string[0] == '0':
            speak += pronounce_number_it(int(string[0])) + " "
            if int(string[1]) == 1:
                speak += "una"  # TODO: valutare forma "l'una"
            else:
                speak += pronounce_number_it(int(string[1]))
        else:
            speak = pronounce_number_it(int(string[0:2]))

        # in italian  "13 e 25"
        speak += " e "

        if string[3:5] == '00':
            speak += "zerozero"
        else:
            if string[3] == '0':
                speak += pronounce_number_it(0) + " "
                speak += pronounce_number_it(int(string[4]))
            else:
                speak += pronounce_number_it(int(string[3:5]))
        return speak
    else:
        if dt.hour == 0 and dt.minute == 0:
            return "mezzanotte"
        if dt.hour == 12 and dt.minute == 0:
            return "mezzogiorno"
        # TODO: "10 e un quarto", "4 e tre quarti" and ot her idiomatic times

        if dt.hour == 0:
            speak = "mezzanotte"
        elif dt.hour == 1 or dt.hour == 13:
            speak = "una"
        elif dt.hour > 13:  # era minore
            speak = pronounce_number_it(dt.hour - 12)
        else:
            speak = pronounce_number_it(dt.hour - 12)

        speak += " e"
        if dt.minute == 0:
            speak = speak[:-2]
            if not use_ampm:
                speak += " in punto"
        else:
            if dt.minute < 10:
                speak += " zero"
            speak += " " + pronounce_number_it(dt.minute)

        if use_ampm:

            if dt.hour < 4:
                speak.strip()
            elif dt.hour > 12:
                speak += " del pomeriggio"
            elif dt.hour > 17:
                speak += " della sera"
            elif dt.hour > 20:
                speak += " della notte"
            else:
                speak += " della mattina"

        return speak



#!/usr/bin/python
from tnalagmes.engines import TNaLaGmesEngine
from tnalagmes.models.objects import Inventory
from tnalagmes.models.agents import Agent
import random


class Scene(TNaLaGmesEngine):
    # TODO basic physics engine, collision detection is the urgent part
    name = "location"
    """
    go up / down / left / right / front /back
    go north / south / west / east ....
    who is in the room
    items in the room
    connected rooms
    describe room
    room _name
    """

    def __init__(self, description="empty room", items=None, npcs=None, directions=None, dimensions=3):
        TNaLaGmesEngine.__init__(self, "scene")
        self.description = description
        self.items = items or Inventory()
        self.npcs = npcs or {}
        self.connections = directions or {}
        self.dimensions = dimensions

    # TODO add object / npc to scene
    def add_connection(self, room, direction="front", message=None):
        if isinstance(room, str):
            room = Scene(room)
        self.connections[direction] = room
        message = message or direction + " there is a " + room.name
        self.description += "\n" + message

    def handle_look(self, intent):
        item = intent.get("item", "room")
        if item == "room":
            return self.description
        return "it is a " + item

    def handle_get(self, intent):
        item = intent.get("item", "nothing")
        if item in self.items:
            item = self.items[item].name
        return "got " + item

    def handle_talk(self, intent):
        npc = intent.get("npc", "yourself")
        if npc in self.npcs:
            npc = self.npcs[npc].name
        return self.talk_to_npc(npc)

    def register_default_intents(self):

        self.register_intent("look", ["look {item}", "look at {item}", "describe {item}"], self.handle_look)
        self.register_intent("get", ["get {item}", "acquire {item}", "fetch {item}", "pick {item}", "stash {item}"],
                             self.handle_look)
        self.register_intent("talk", ["talk with {npc}", "engage {npc}", "interact with {npc}"],
                             self.handle_look)

    def get_item(self, item):
        if item in self.items:
            item = self.items[item]
            self.items.pop(item)
            return item
        return None

    def talk_to_npc(self, npc, utterance="hello"):
        if npc in self.npcs:
            return self.npcs[npc].parse_command(utterance)
        return "talk to who?"


class World(Scene):
    name = "Text Universe"
    """
    volume / height / width / area / dimensions
    number of rooms
    teleport to room
    how many players
    how many npcs
    how many items
    create scene
    create npc
    """

    def __init__(self, description="the text universe", agent=None, scenes=None, dimensions=3):
        Scene.__init__(self, dimensions=dimensions)
        self.object_type = "world"
        self.agent = agent or Agent()
        self.description = description
        self.scenes = scenes or {"emptiness": Scene()}

    def register_default_intents(self):
        self.register_intent("up", ["up"], self.handle_up)
        self.register_intent("down", ["down"], self.handle_down)
        self.register_intent("front", ["forward"], self.handle_front)
        self.register_intent("back", ["back", "backward"], self.handle_back)
        self.register_intent("left", ["left"], self.handle_left)
        self.register_intent("right", ["right"], self.handle_right)
        self.register_intent("north", ["north"], self.handle_north)
        self.register_intent("south", ["south"], self.handle_south)
        self.register_intent("east", ["east"], self.handle_east)
        self.register_intent("west", ["west"], self.handle_west)
        self.register_intent("northeast", ["northeast"], self.handle_northeast)
        self.register_intent("northwest", ["northwest"], self.handle_northwest)
        self.register_intent("southeast", ["southeast"], self.handle_southeast)
        self.register_intent("southwest", ["southwest"], self.handle_southweast)
        self.register_intent("describe", ["describe room", "describe surroundings", "look"], self.handle_describe)

    def place_player(self, scene_name=None):
        scene_name = scene_name or random.choice(self.scenes)
        self.agent.scene = self.scenes.get(scene_name)

    @property
    def current_scene(self):
        return self.agent.scene

    def handle_describe(self, intent):
        """

        :param intent:
        :return:
        """
        return self.agent.scene.description

    def handle_up(self, intent):
        return ""

    def handle_down(self, intent):
        return ""

    def handle_front(self, intent):
        return ""

    def handle_back(self, intent):
        return ""

    def handle_left(self, intent):
        return ""

    def handle_right(self, intent):
        return ""

    def handle_north(self, intent):
        return ""

    def handle_south(self, intent):
        return ""

    def handle_east(self, intent):
        return ""

    def handle_west(self, intent):
        return ""

    def handle_northeast(self, intent):
        return ""

    def handle_northwest(self, intent):
        return ""

    def handle_southeast(self, intent):
        return ""

    def handle_southweast(self, intent):
        return ""


class PhysicsEngine(World):
    # TODO real physics simulation helpers
    laws = {}



# -*- coding: utf-8 -*-
#
# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Format functions for french (fr)

"""

from tnalagmes.lang.format_common import convert_to_mixed_fraction

NUM_STRING_FR = {
    0: 'zéro',
    1: 'un',
    2: 'deux',
    3: 'trois',
    4: 'quatre',
    5: 'cinq',
    6: 'six',
    7: 'sept',
    8: 'huit',
    9: 'neuf',
    10: 'dix',
    11: 'onze',
    12: 'douze',
    13: 'treize',
    14: 'quatorze',
    15: 'quinze',
    16: 'seize',
    20: 'vingt',
    30: 'trente',
    40: 'quarante',
    50: 'cinquante',
    60: 'soixante',
    70: 'soixante-dix',
    80: 'quatre-vingt',
    90: 'quatre-vingt-dix'
}

FRACTION_STRING_FR = {
    2: 'demi',
    3: 'tiers',
    4: 'quart',
    5: 'cinquième',
    6: 'sixième',
    7: 'septième',
    8: 'huitième',
    9: 'neuvième',
    10: 'dixième',
    11: 'onzième',
    12: 'douzième',
    13: 'treizième',
    14: 'quatorzième',
    15: 'quinzième',
    16: 'seizième',
    17: 'dix-septième',
    18: 'dix-huitième',
    19: 'dix-neuvième',
    20: 'vingtième'
}


def nice_number_fr(number, speech, denominators):
    """ French helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 et demi" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    strNumber = ""
    whole = 0
    num = 0
    den = 0

    result = convert_to_mixed_fraction(number, denominators)

    if not result:
        # Give up, just represent as a 3 decimal number
        whole = round(number, 3)
    else:
        whole, num, den = result

    if not speech:
        if num == 0:
            strNumber = '{:,}'.format(whole)
            strNumber = strNumber.replace(",", " ")
            strNumber = strNumber.replace(".", ",")
            return strNumber
        else:
            return '{} {}/{}'.format(whole, num, den)
    else:
        if num == 0:
            # if the number is not a fraction, nothing to do
            strNumber = str(whole)
            strNumber = strNumber.replace(".", ",")
            return strNumber
        den_str = FRACTION_STRING_FR[den]
        # if it is not an integer
        if whole == 0:
            # if there is no whole number
            if num == 1:
                # if numerator is 1, return "un demi", for example
                strNumber = 'un {}'.format(den_str)
            else:
                # else return "quatre tiers", for example
                strNumber = '{} {}'.format(num, den_str)
        elif num == 1:
            # if there is a whole number and numerator is 1
            if den == 2:
                # if denominator is 2, return "1 et demi", for example
                strNumber = '{} et {}'.format(whole, den_str)
            else:
                # else return "1 et 1 tiers", for example
                strNumber = '{} et 1 {}'.format(whole, den_str)
        else:
            # else return "2 et 3 quart", for example
            strNumber = '{} et {} {}'.format(whole, num, den_str)
        if num > 1 and den != 3:
            # if the numerator is greater than 1 and the denominator
            # is not 3 ("tiers"), add an s for plural
            strNumber += 's'

    return strNumber


def pronounce_number_fr(num, places=2):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'cinq virgule deux'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
    Returns:
        (str): The pronounced number
    """
    if abs(num) >= 100:
        # TODO: Support for numbers over 100
        return str(num)

    result = ""
    if num < 0:
        result = "moins "
    num = abs(num)

    if num > 16:
        tens = int(num - int(num) % 10)
        ones = int(num - tens)
        if ones != 0:
            if tens > 10 and tens <= 60 and int(num - tens) == 1:
                result += NUM_STRING_FR[tens] + "-et-" + NUM_STRING_FR[ones]
            elif num == 71:
                result += "soixante-et-onze"
            elif tens == 70:
                result += NUM_STRING_FR[60] + "-"
                if ones < 7:
                    result += NUM_STRING_FR[10 + ones]
                else:
                    result += NUM_STRING_FR[10] + "-" + NUM_STRING_FR[ones]
            elif tens == 90:
                result += NUM_STRING_FR[80] + "-"
                if ones < 7:
                    result += NUM_STRING_FR[10 + ones]
                else:
                    result += NUM_STRING_FR[10] + "-" + NUM_STRING_FR[ones]
            else:
                result += NUM_STRING_FR[tens] + "-" + NUM_STRING_FR[ones]
        else:
            if num == 80:
                result += "quatre-vingts"
            else:
                result += NUM_STRING_FR[tens]
    else:
        result += NUM_STRING_FR[int(num)]

    # Deal with decimal part
    if not num == int(num) and places > 0:
        result += " virgule"
        place = 10
        while int(num * place) % 10 > 0 and places > 0:
            result += " " + NUM_STRING_FR[int(num * place) % 10]
            place *= 10
            places -= 1
    return result


def nice_time_fr(dt, speech=True, use_24hour=False, use_ampm=False):
    """
    Format a time to a comfortable human format

    For example, generate 'cinq heures trente' for speech or '5:30' for
    text display.

    Args:
        dt (datetime): date to format (assumes already in local timezone)
        speech (bool): format for speech (default/True) or display (False)=Fal
        use_24hour (bool): output in 24-hour/military or 12-hour format
        use_ampm (bool): include the am/pm for 12-hour format
    Returns:
        (str): The formatted time string
    """
    if use_24hour:
        # e.g. "03:01" or "14:22"
        string = dt.strftime("%H:%M")
    else:
        if use_ampm:
            # e.g. "3:01 AM" or "2:22 PM"
            string = dt.strftime("%I:%M %p")
        else:
            # e.g. "3:01" or "2:22"
            string = dt.strftime("%I:%M")
        if string[0] == '0':
            string = string[1:]  # strip leading zeros

    if not speech:
        return string

    # Generate a speakable version of the time
    speak = ""
    if use_24hour:

        # "13 heures trente"
        if dt.hour == 0:
            speak += "minuit"
        elif dt.hour == 12:
            speak += "midi"
        elif dt.hour == 1:
            speak += "une heure"
        else:
            speak += pronounce_number_fr(dt.hour) + " heures"

        if dt.minute != 0:
            speak += " " + pronounce_number_fr(dt.minute)

    else:
        # Prepare for "trois heures moins le quart"
        if dt.minute == 35:
            minute = -25
            hour = dt.hour + 1
        elif dt.minute == 40:
            minute = -20
            hour = dt.hour + 1
        elif dt.minute == 45:
            minute = -15
            hour = dt.hour + 1
        elif dt.minute == 50:
            minute = -10
            hour = dt.hour + 1
        elif dt.minute == 55:
            minute = -5
            hour = dt.hour + 1
        else:
            minute = dt.minute
            hour = dt.hour

        if hour == 0:
            speak += "minuit"
        elif hour == 12:
            speak += "midi"
        elif hour == 1 or hour == 13:
            speak += "une heure"
        elif hour < 13:
            speak = pronounce_number_fr(hour) + " heures"
        else:
            speak = pronounce_number_fr(hour - 12) + " heures"

        if minute != 0:
            if minute == 15:
                speak += " et quart"
            elif minute == 30:
                speak += " et demi"
            elif minute == -15:
                speak += " moins le quart"
            else:
                speak += " " + pronounce_number_fr(minute)

        if use_ampm:
            if hour > 17:
                speak += " du soir"
            elif hour > 12:
                speak += " de l'après-midi"
            elif hour > 0 and hour < 12:
                speak += " du matin"

    return speak



from tnalagmes.engines.oregon75 import Oregon75Engine
from tnalagmes.data.template_data import GAME_EVENTS, RANDOM_EVENTS, TERMINOLOGY


class TemplateGame(Oregon75Engine):
    DATA = GAME_EVENTS
    RANDOM_EVENTS = RANDOM_EVENTS
    TERMINOLOGY = TERMINOLOGY

    def __init__(self):
        Oregon75Engine.__init__(self)


if __name__ == "__main__":
    game = TemplateGame()
    game.run()



TERMINOLOGY = {
    'ammunition': ['ammunition'],
    'armour': ['clothing'],
    'attack_damage': ['injuries'],
    'enemy': ['zombies'],
    'fuel': ['generator fuel'],
    'medicine': ['antidote'],
    'passive_damage': ['z-virus'],
    'supplies': ['food'],
    'currency': ['inventory slots']
}

RANDOM_EVENTS = {
    'attack': {'conclusion': ["just barely, but you managed to defeat it",
                              "headshot!",
                              "solved with an axe to the head"],
               'damage': 'received damage',
               'die': '',
               'error': 'you trapped the zombie, but are are now running out of ammunition',
               'events': [],
               'intro': 'a zombie sneaks on you'},
    'illness': {'conclusion': '',
                'damage': '',
                'die': '',
                'error': '',
                'events': [],
                'intro': 'you are starving'},
    'poison': {'conclusion': '',
               'die': 'you die since you have no medicine',
               'error': '',
               'events': [],
               'intro': 'you ate infected food'},
    'animal': {'conclusion': 'scared them away',
               'damage': 'got bitten',
               'die': '',
               'error': ["lost a boot", "lost clothing", "lost knife", "lost food", "damaged jacket"],
               'events': [],
               'intro': 'rabid dogs attack!'},
    'companion': {'conclusion': '',
                  'damage': '',
                  'die': '',
                  'error': '',
                  'events': [],
                  'intro': ['you spend time making zombie traps',
                            'a zombie managed to bypass your barricade, you killed and spent time looking for the hole',
                            'barbed wire fence maintenance']},
    'bad_terrain': {'conclusion': 'you use food and clothes to make a decoy and lure them away',
                    'intro': 'a zombie horde is gathering outside'},
    'companion_lose': {'conclusion': 'spent time repairing it',
                       'intro': 'the barricade sustained damage'},
    'find_supplies': {'conclusion': '', 'intro': 'found supplies'},
    'fuel_damage': {'conclusion': 'spend time scavenging for pieces',
                    'intro': 'generator problems'},
    'shelter_damage': {'conclusion': 'lose time and supplies fixing it',
                       'intro': ['shelter damage', 'barricade suffered damage']},
    'shelter_fire': {'conclusion': 'supplies damaged',
                     'intro': ['kitchen fire']},
    'supply_damage': {'conclusion': 'lose time scavenging for more',
                      'intro': 'supplies perished'},
    'vehicle_damage': {'conclusion': 'lose time, you must stay at home until they leave',
                       'intro': 'zombie swarm at the door'},
    'rain': {'conclusion': 'time and supplies lost',
             'intro': 'rain'
             },
    'storm': {'conclusion': 'supplies damaged', 'intro': 'storm'},
    'fog': {'conclusion': 'time is lost',
            'intro': 'lose your way in heavy fog'},
    'cold': {'conclusion': 'you have enough clothing',
             'error': 'frostbite',
             'intro': 'cold'}
}

GAME_EVENTS = {
    'damage': {'conclusion': '',
               'error': 'you ran out medical supplies\nyou died of',
               'high': 'critical health --medicine used',
               'intro': '',
               'mild': 'low health ---medicine used'},
    'difficulty_damage': {'conclusion': '',
                          'error': '',
                          'events': [],
                          'intro': 'time and supplies lost'},
    'easy_difficulty': {'intro': 'you made it safely to the lab',
                        'conclusion': 'you have no idea how to break in',
                        'events': ['you got lost---lose valuable time trying to '
                                   'find trail!',
                                   'vehicle damaged!---lose time and supplies',
                                   'dangerous path ahead']},
    'enemy_encounter': {'conclusion': '',
                        'damage': 'sustained damage',
                        'die': 'got killed by ',
                        'enemy_run': 'they did not attack',
                        'high_score': 'you killed them all',
                        'hostile_conclusion': ' attacked you',
                        'hostile_intro': 'they look hostile',
                        'intro': ' ahead.',
                        'low_score': 'kinda slow with your shooting',
                        'number_questions': ['(1) run  (2) attack  (3) continue  '
                                             '(4) circle'],
                        'peaceful_conclusion': ' couldn\'t reach you',
                        'peaceful_intro': "they appear crippled"},
    'explore': {'conclusion': '',
                'error': 'can not explore without ammunition',
                'events': ['found supplies!',
                           'sorry---no luck today',
                           'found a lot of supplies'],
                'intro': ''},
    'game_over': {'conclusion': 'distance to lab: ',
                  'error': 'time has run out',
                  'intro': 'you never found the cure'},
    'hard_difficulty': {'conclusion': 'violent thoughts plague your mind',
                        'intro': 'You notice a mirror, your eyes are turning red'},
    'heal': {'conclusion': "you prepare 20 vials of virus retardant",
             'die': 'you died of ',
             'error': "you are out of antidote",
             'events': [],
             'intro': ''},
    'intro': {
        'conclusion': 'you hear rumours about a cure being successfully developed in a lab just before all communication broke down...\nfind the cure before you turn',
        'intro': 'It\'s been 1 year since the outbreak, yesterday you got bitten, damn zombie found a hole in your defences...'},
    'inventory': {'conclusion': 'after all your choices, you now have '
                                '{inv.money} slots left',
                  'error': 'you overshoot, you only had {inv.start_money} slots to '
                           'spend.  chose again',
                  'intro': 'how many slots do you want to assign to '},
    'lose': {'intro': 'you died, just before you turned you saw your first face in months, you tried to say something',
             "conclusion": "brains..."},
    'low_supplies': {'intro': 'low supplies'},
    'maintenance': {'conclusion': '',
                    'error': 'not enough supplies',
                    'intro': 'consume supplies (1) a little  (2) moderately (3) a lot'},
    'medium_difficulty': {'conclusion': 'you begin to wonder if the cure is real', 'intro': 'You break into the lab'},
    'numeric': {'error': 'impossible', 'high': 'too high', 'low': 'too low'},
    'shop': {'conclusion': '',
             'error': "you don't have that much slots your spending down",
             'events': [],
             'intro': 'what are we scavenging for'},
    'turn': {'die': 'you ran out of supplies and starved to death',
             'intro': '(1) go scavenging, (2) explore, (3) travel'},
'win': {'intro': 'you made it to the lab\nyou found the cure on time, however you are no scientist\nyou cure yourself and use every last bit of the ZVirus cure\n you realize the world is still full of zombies...\nhopefully you are now immune, happy zombie hunting!'}}
